#!/usr/bin/env python
__doc__ = """

Read-only/writable TensorData classes.

Kisuk Lee <kisuklee@mit.edu>, 2015-2016
"""

import math
import numpy as np

from box import *
from vector import *

class TensorData(object):
    """Read-only tensor data.

    The 1st dimension is regarded as parallel channels, and arbitrary access
    along this dimension is not allowed. Threfore, every data access should be
    made through 3D vector, not 4D.

    Attributes:
        _data:   numpy 4D array (channel,z,y,x)
        _dim:    Dimension of each channel
        _offset: Coordinate offset from the origin
        _bb:     Bounding box
        _fov:    Patch size
        _rg:     Range (update dep.: dim, offset, fov)
    """

    def __init__(self, data, fov=(0,0,0), offset=(0,0,0)):
        """
        Initialize a TensorData object.
        """
        # Set immutable attributes.
        self._data   = self._check_data(data)
        self._dim    = Vec3d(self._data.shape[1:])
        self._offset = Vec3d(offset)

        # Set bounding box.
        self._bb = Box((0,0,0), self._dim)
        self._bb.translate(self._offset)

        # Set fov (patch size).
        self.set_fov(fov)

    def set_fov(self, fov):
        """
        Set a nonnegative field of view (FoV), i.e., patch size.
        """
        # Zero FoV indicates covering the whole volume.
        fov = Vec3d(fov)
        if fov == (0,0,0):
            fov = Vec3d(self._dim)

        # FoV should be nonnegative, and smaller than data dimension.
        assert fov==minimum(maximum(fov,(0,0,0)), self._dim)

        # Set FoV.
        self._fov = fov

        # Update range.
        self._set_range()

    def get_patch(self, pos):
        """
        Extract a patch of size _fov centered on pos.
        """
        # Check validity.
        assert self._rg.contains(pos)
        # Local coordinate system
        loc  = pos - self._offset
        box  = centered_box(loc, self._fov)
        vmin = box.min()
        vmax = box.max()
        return np.copy(self._data[:,vmin[0]:vmax[0],
                                    vmin[1]:vmax[1],
                                    vmin[2]:vmax[2]])

    ####################################################################
    ## Public methods for accessing attributes
    ####################################################################

    def get_data(self):
        return self._data

    def shape(self):
        """Return data shape (c,z,y,x)."""
        return self._data.shape

    def dim(self):
        """Return channel shape (z,y,x)."""
        return Vec3d(self._dim)

    def fov(self):
        return Vec3d(self._fov)

    def offset(self):
        return Vec3d(self._offset)

    def bounding_box(self):
        return Box(self._bb)

    def range(self):
        return Box(self._rg)

    ####################################################################
    ## Private helper methods.
    ####################################################################

    def _check_data(self, data):
        # Data should be either numpy 3D or 4D array.
        assert isinstance(data, np.ndarray)
        assert data.ndim==3 or data.ndim==4

        # Add channel dimension if data is 3D array.
        if data.ndim == 3:
            data = data[np.newaxis,...]

        return data

    def _set_range(self):
        """Set a valid range for extracting patches."""
        top  = self._fov/2                # Top margin
        btm  = self._fov - top - (1,1,1)  # Bottom margin
        vmin = self._offset + top
        vmax = self._offset + self._dim - btm

        self._rg = Box(vmin, vmax)

    # String representaion (for printing and debugging).
    def __str__( self ):
        return "<TensorData>\nshape: %s\ndim: %s\nFoV: %s\noffset: %s\n" % \
               (self.shape(), self._dim, self._fov, self._offset)


class WritableTensorData(TensorData):
    """
    Writable tensor data.
    """

    def __init__(self, data_or_shape, fov=(0,0,0), offset=(0,0,0)):
        """
        Initialize a writable tensor data, or create a new tensor of zeros.
        """
        if isinstance(data_or_shape, np.ndarray):
            TensorData.__init__(self, data_or_shape, fov, offset)
        else:
            data = np.zeros(data_or_shape, dtype='float32')
            TensorData.__init__(self, data, fov, offset)

    def set_patch(self, pos, patch, op=None):
        """
        Write a patch of size _fov centered on pos.
        """
        assert self._rg.contains(pos)
        patch = self._check_data(patch)
        dim = patch.shape[1:]
        assert dim==self._fov
        box = centered_box(pos, dim)

        # Local coordinate
        box.translate(-self._offset)
        vmin = box.min()
        vmax = box.max()
        lval = 'self._data[:,vmin[0]:vmax[0],vmin[1]:vmax[1],vmin[2]:vmax[2]]'
        rval = 'patch'
        if op is None:
            exec '{}={}'.format(lval, rval)
        else:
            exec '{}={}({},{})'.format(lval, op, lval, rval)


class WritableTensorDataWithMask(WritableTensorData):
    """
    Writable tensor data with blending mask.
    """

    def __init__(self, data_or_shape, fov=(0,0,0), offset=(0,0,0)):
        """
        Initialize a writable tensor data, or create a new tensor of zeros.
        """
        WritableTensorData.__init__(self, data_or_shape, fov, offset)

        # Set norm.
        self._norm = WritableTensorData(self.shape(), fov, offset)

    def set_patch(self, pos, patch, op='np.add', mask=None):
        """
        Write a patch of size _fov centered on pos.
        """
        # Default mask.
        if mask is None:
            mask = np.ones(patch.shape, dtype='float32')
        # Set patch.
        WritableTensorData.set_patch(self, pos, mask*patch, op)
        # Set normalization.
        self._norm.set_patch(pos, mask, op='np.add')

    def get_norm(self):
        return self._norm._data

    def get_data(self):
        # return self._data/self._norm._data
        # Temporary in-place normalization.
        self._data = np.divide(self._data, self._norm._data, self._data)
        return self._data

    def get_unnormalized_data(self):
        return WritableTensorData.get_data(self)


########################################################################
## Unit Testing
########################################################################
if __name__ == "__main__":

    import unittest

    ####################################################################
    class UnitTestTensorData(unittest.TestCase):

        def setup(self):
            pass

        def testCreation(self):
            data = np.zeros((4,4,4,4))
            T = TensorData(data, (3,3,3), (1,1,1))
            self.assertTrue(T.shape()==(4,4,4,4))
            self.assertTrue(T.offset()==(1,1,1))
            self.assertTrue(T.fov()==(3,3,3))
            bb = T.bounding_box()
            rg = T.range()
            self.assertTrue(bb==Box((1,1,1),(5,5,5)))
            self.assertTrue(rg==Box((2,2,2),(4,4,4)))

        def testGetPatch(self):
            # (4,4,4) random 3D araray
            data = np.random.rand(4,4,4)
            T = TensorData(data, (3,3,3))
            p = T.get_patch((2,2,2))
            self.assertTrue(np.array_equal(data[1:,1:,1:], p[0, ...]))
            T.set_fov((2,2,2))
            p = T.get_patch((2,2,2))
            self.assertTrue(np.array_equal(data[1:3,1:3,1:3], p[0, ...]))


    ####################################################################
    class UnitTestWritableTensorData(unittest.TestCase):

        def setup(self):
            pass

        def testCreation(self):
            data = np.zeros((1,4,4,4))
            T = WritableTensorData(data, (3,3,3), (1,1,1))
            self.assertTrue(T.shape()==(1,4,4,4))
            self.assertTrue(T.offset()==(1,1,1))
            self.assertTrue(T.fov()==(3,3,3))
            bb = T.bounding_box()
            rg = T.range()
            self.assertTrue(bb==Box((1,1,1),(5,5,5)))
            self.assertTrue(rg==Box((2,2,2),(4,4,4)))

        def testSetPatch(self):
            T = WritableTensorData(np.zeros((1,5,5,5)), (3,3,3), (1,1,1))
            p = np.random.rand(1,3,3,3)
            self.assertFalse(np.array_equal(p, T.get_patch((4,4,4))))
            T.set_patch((4,4,4), p)
            self.assertTrue(np.array_equal(p, T.get_patch((4,4,4))))

    ####################################################################
    unittest.main()

    ####################################################################



#!/usr/bin/env python
__doc__ = """

Inference outputs.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import numpy as np

import box
from tensor import WritableTensorData as WTD, WritableTensorDataWithMask as WTDM

def prepare_outputs(spec, locs, blend=False, blend_mode=''):
    blend_pool = ['','bump']
    b = blend_mode.lower()
    if b not in blend_pool:
        raise RuntimeError('unknown output blend type [%s]' % b)
    if b == '':
        b = 'Blend'
    else:
        b = b[0].capitalize() + b[1:] + 'Blend'
    outputs = eval(b + '(spec, locs, blend)')
    return outputs


class Blend(object):
    """
    Blend interface.
    """

    def __init__(self, spec, locs, blend=False):
        """Initialize Blend."""
        self.spec  = spec
        self.locs  = locs
        self.blend = blend
        self._prepare_data()

    def push(self, loc, sample):
        """Write to data."""
        for k, v in sample.iteritems():
            assert k in self.data
            self.data[k].set_patch(loc, v, op=self.op)

    def get_data(self, key):
        """Get inference output data."""
        assert key in self.data
        return self.data[key].get_data()

    ####################################################################
    ## Private Methods.
    ####################################################################

    def _prepare_data(self):
        """
        TODO(kisuk): Documentation.
        """
        assert len(self.locs) > 0
        rmin = self.locs[0]
        rmax = self.locs[-1]

        self.data = dict()
        self.op   = None
        for k, v in self.spec.iteritems():
            fov = v[-3:]
            a = box.centered_box(rmin, fov)
            b = box.centered_box(rmax, fov)
            c = a.merge(b)
            shape = v[:-3] + tuple(c.size())
            # Inference with overlapping window.
            if self.blend:
                self.data[k] = WTDM(shape, fov, c.min())
                self.op = 'np.add'
            else:
                self.data[k] = WTD(shape, fov, c.min())


class BumpBlend(Blend):
    """
    Blending with bump function.
    """

    def __init__(self, spec, locs, blend=False):
        """Initialize BumpBlend."""
        Blend.__init__(self, spec, locs, blend)

        # Inference with overlapping window.
        self.max_logits = None
        if blend:
            max_logits = dict()
            # Compute max_logit for numerical stability.
            for k, v in self.data.iteritems():
                fov = tuple(v.fov())
                data = np.zeros(v.dim())
                data.fill(-np.inf)
                max_logit = WTD(data, fov, v.offset())
                max_logit_window = self._bump_logit_map(fov)
                for loc in self.locs:
                    max_logit.set_patch(loc, max_logit_window, op='np.maximum')
                max_logits[k] = max_logit
            self.max_logits = max_logits

    def push(self, loc, sample):
        """Blend with data."""
        for k, v in sample.iteritems():
            assert k in self.data
            mask = self._get_mask(k, loc)
            self.data[k].set_patch(loc, v, op=self.op, mask=mask)

    ####################################################################
    ## Private methods.
    ####################################################################

    def _get_mask(self, key, loc):
        mask = None
        if self.blend:
            assert key in self.max_logits
            max_logit = self.max_logits[key].get_patch(loc)
            mask = self._bump_map(max_logit.shape[-3:], max_logit[0,...])
        return mask

    def _bump_logit(self, z, y, x, t=1.5):
        return -(x*(1-x))**(-t)-(y*(1-y))**(-t)-(z*(1-z))**(-t)

    def _bump_logit_map(self, dim):
        x = range(dim[-1])
        y = range(dim[-2])
        z = range(dim[-3])
        zv, yv, xv = np.meshgrid(z, y, x, indexing='ij')
        xv = (xv+1.0)/(dim[-1]+1.0)
        yv = (yv+1.0)/(dim[-2]+1.0)
        zv = (zv+1.0)/(dim[-3]+1.0)
        return self._bump_logit(zv, yv, xv)

    def _bump_map(self, dim, max_logit):
        return np.exp(self._bump_logit_map(dim) - max_logit)



#!/usr/bin/env python
__doc__ = """

Utility functions.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import numpy as np

def check_volume(data):
    """Ensure that data is numpy 3D array."""
    assert isinstance(data, np.ndarray)

    if data.ndim == 2:
        data = data[np.newaxis,...]
    elif data.ndim == 3:
        pass
    elif data.ndim == 4:
        assert data.shape[0]==1
        data = np.reshape(data, data.shape[-3:])
    else:
        raise RuntimeError('data must be a numpy 3D array')

    assert data.ndim==3
    return data


def check_tensor(data):
    """Ensure that data is numpy 4D array."""
    assert isinstance(data, np.ndarray)

    if data.ndim == 2:
        data = data[np.newaxis,np.newaxis,...]
    elif data.ndim == 3:
        data = data[np.newaxis,...]
    elif data.ndim == 4:
        pass
    else:
        raise RuntimeError('data must be a numpy 4D array')

    assert data.ndim==4
    return data


def fill_data(shape, filler={'type':'zero'}, dtype='float32'):
    """
    Return numpy array of shape, filled with specified values.

    Args:
        shape: Array shape.
        filler: {'type':'zero'} (default)
                {'type':'one'}
                {'type':'constant', 'value':%f}
                {'type':'gaussian', 'loc':%f, 'scale':%f}
                {'type':'uniform', 'low':%f, 'high':%f}
                {'type':'randi', 'low':%d, 'high':%d}

    Returs:
        data: Numpy array of shape, filled with specified values.
    """
    data = np.zeros(shape, dtype=dtype)

    assert 'type' in filler
    if filler['type'] == 'zero':
        # Fill zeros.
        pass
    elif filler['type'] == 'one':
        # Fill ones.
        data = np.ones(shape, dtype=dtype)
    elif filler['type'] == 'constant':
        # Fill constant value.
        assert 'value' in filler
        data[:] = filler['value']
    elif filler['type'] == 'gaussian':
        # Fill random numbers from Gaussian(loc, scale).
        loc = filler.get('mean', 0.0)
        scale = filler.get('std', 1.0)
        data[:] = np.random.normal(loc=loc, scale=scale, size=shape)
    elif filler['type'] == 'uniform':
        # Fill random numbers from Uniform(low, high).
        low = filler.get('low', 0.0)
        high = filler.get('high', 1.0)
        data[:] = np.random.uniform(low=low, high=high, size=shape)
    elif filler['type'] == 'randint':
        low = filler.get('low', 0)
        high = filler.get('high', None)
        data[:] = np.random.randint(low=low, high=high, size=shape)
    else:
        raise RuntimeError('invalid filler type [%s]' % filler['type'])

    return data



#!/usr/bin/env python
__doc__ = """

Warp data augmentation.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

from box import Box
import data_augmentation
from warping import _warping
import numpy as np
from utils import check_tensor, check_volume
from vector import Vec3d

class WarpAugment(data_augmentation.DataAugment):
    """
    Warp.
        1. Continuous rotation
        2. Shear
        3. Twist
        4. Scale
        5. Perspective stretch
    """

    def __init__(self, skip_ratio=0.3):
        """Initialize WarpAugment."""
        self.ratio = skip_ratio

        # DEBUG
        # self.count = dict(skip=0, warp=0)
        # self.counter = 0

    def prepare(self, spec, **kwargs):
        """
        Randomly draw warp parameters and compute required (mostly
        larger than original) image sizes.
        """
        # Skip.
        self.skip = False
        if self.ratio > np.random.rand():
            self.skip = True
            return dict(spec)

        imgs = kwargs['imgs']

        # Compute the largest image size.
        b = Box((0,0,0), (0,0,0))  # Empty box.
        for k, v in spec.iteritems():
            b = b.merge(Box((0,0,0), v[-3:]))
        maxsz = tuple(b.size())

        # Randomly draw warp parameters.
        # TODO(kisuk): Optional parameter 'amount'?
        params = warping.getWarpParams(maxsz, **kwargs)
        self.size = tuple(x for x in params[0])  # Convert to tuple.
        size_diff = tuple(x - y for x, y in zip(self.size,maxsz))
        self.rot     = params[1]
        self.shear   = params[2]
        self.scale   = params[3]
        self.stretch = params[4]
        self.twist   = params[5]

        # Save original spec.
        self.spec = dict(spec)

        # Replace every shape to the largest required one.
        # TODO(kisuk): Is this correct?
        ret = dict()
        for k, v in spec.iteritems():
            if k in imgs:  # Images.
                ret[k] = v[:-3] + self.size
            else:  # Labels and masks.
                ret[k] = v[:-3] + tuple(x + y for x, y in zip(v[-3:],size_diff))
        return ret

    def augment(self, sample, **kwargs):
        """Apply warp data augmentation."""
        # DEBUG
        # print '\n[WarpAugment]'
        # self.counter += 1
        # if self.skip:
        #     self.count['skip'] += 1
        # else:
        #     self.count['warp'] += 1
        # for k,v in self.count.iteritems():
        #     print '{}={}'.format(k,'%0.3f'%(v/float(self.counter)))

        if self.skip:
            return sample

        # DEBUG
        #print 'rot      = {}'.format(self.rot)
        #print 'shear    = {}'.format(self.shear)
        #print 'scale    = {}'.format(self.scale)
        #print 'stretch  = {}'.format(self.stretch)
        #print 'twist    = {}'.format(self.twist)
        #print 'req_size = {}'.format(self.size)

        imgs = kwargs['imgs']

        # Apply warp to each tensor.
        for k, v in sample.iteritems():
            v = check_tensor(v)
            v = np.transpose(v, (1,0,2,3))
            if k in imgs:  # Images.
                v = warping.warp3d(v, self.spec[k][-3:],
                    self.rot, self.shear, self.scale, self.stretch, self.twist)
            else:  # Labels and masks.
                v = warping.warp3dLab(v, self.spec[k][-3:], self.size,
                    self.rot, self.shear, self.scale, self.stretch, self.twist)
            sample[k] = np.transpose(v, (1,0,2,3))
        return sample


if __name__ == "__main__":

    # Fov.
    spec = dict()
    spec['input/p3'] = (5,109,109)
    spec['input/p2'] = (7,73,73)
    spec['input/p1'] = (9,45,45)
    spec['label']    = (3,1,1,1)

    # In/out size.
    outsz = Vec3d(5,100,100)
    for k, v in spec.iteritems():
        newv = tuple(Vec3d(v[-3:]) + outsz - Vec3d(1,1,1))
        spec[k] = v[:-3] + newv

    # Augmentation.
    aug = WarpAugment()

    # Test.
    ret = aug.prepare(spec, imgs=['input/p3','input/p2','input/p1'])
    print ret
    print aug.spec
    print aug.rot
    print aug.shear
    print aug.scale
    print aug.stretch
    print aug.twist



#!/usr/bin/env python
__doc__ = """

3D Vector class.

Modified from the vec3d class on the pygame wiki site.
http://pygame.org/wiki/3DVectorClass

Kisuk Lee <kisuklee@mit.edu>, 2015-2016
"""

import math
import operator

class Vec3d(object):
    """
    3d vector class, supports vector and scalar operators,
    and also provides a bunch of high level functions.
    reproduced from the vec2d class on the pygame wiki site.

    """
    __slots__ = ['x', 'y', 'z']

    def __init__(self, x_or_triple, y = None, z = None):
        if y == None:
            self.x = x_or_triple[0]
            self.y = x_or_triple[1]
            self.z = x_or_triple[2]
        else:
            self.x = x_or_triple
            self.y = y
            self.z = z

    def __len__(self):
        return 3

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        elif key == 2:
            return self.z
        else:
            raise IndexError("Invalid subscript "+str(key)+" to Vec3d")

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        elif key == 1:
            self.y = value
        elif key == 2:
            self.z = value
        else:
            raise IndexError("Invalid subscript "+str(key)+" to Vec3d")

    # String representaion (for debugging)
    def __repr__(self):
        return 'Vec3d(%s, %s, %s)' % (self.x, self.y, self.z)

    # Comparison
    def __eq__(self, v):
        if hasattr(v, "__getitem__") and len(v) == 3:
            return self.x == v[0] and self.y == v[1] and self.z == v[2]
        else:
            return False

    def __ne__(self, v):
        if hasattr(v, "__getitem__") and len(v) == 3:
            return self.x != v[0] or self.y != v[1] or self.z != v[2]
        else:
            return True

    def __nonzero__(self):
        return self.x or self.y or self.z

    # Generic operator handlers
    def _o2(self, v, f):
        "Any two-operator operation where the left operand is a Vec3d"
        if isinstance(v, Vec3d):
            return Vec3d(f(self.x, v.x),
                         f(self.y, v.y),
                         f(self.z, v.z))
        elif (hasattr(v, "__getitem__")):
            return Vec3d(f(self.x, v[0]),
                         f(self.y, v[1]),
                         f(self.z, v[2]))
        else:
            return Vec3d(f(self.x, v),
                         f(self.y, v),
                         f(self.z, v))

    def _r_o2(self, v, f):
        "Any two-operator operation where the right operand is a Vec3d"
        if (hasattr(v, "__getitem__")):
            return Vec3d(f(v[0], self.x),
                         f(v[1], self.y),
                         f(v[2], self.z))
        else:
            return Vec3d(f(v, self.x),
                         f(v, self.y),
                         f(v, self.z))

    def _io(self, v, f):
        "inplace operator"
        if (hasattr(v, "__getitem__")):
            self.x = f(self.x, v[0])
            self.y = f(self.y, v[1])
            self.z = f(self.z, v[2])
        else:
            self.x = f(self.x, v)
            self.y = f(self.y, v)
            self.z = f(self.z, v)
        return self

    # Addition
    def __add__(self, v):
        if isinstance(v, Vec3d):
            return Vec3d(self.x + v.x, self.y + v.y, self.z + v.z)
        elif hasattr(v, "__getitem__"):
            return Vec3d(self.x + v[0], self.y + v[1], self.z + v[2])
        else:
            return Vec3d(self.x + v, self.y + v, self.z + v)
    __radd__ = __add__

    def __iadd__(self, v):
        if isinstance(v, Vec3d):
            self.x += v.x
            self.y += v.y
            self.z += v.z
        elif hasattr(v, "__getitem__"):
            self.x += v[0]
            self.y += v[1]
            self.z += v[2]
        else:
            self.x += v
            self.y += v
            self.z += v
        return self

    # Subtraction
    def __sub__(self, v):
        if isinstance(v, Vec3d):
            return Vec3d(self.x - v.x, self.y - v.y, self.z - v.z)
        elif (hasattr(v, "__getitem__")):
            return Vec3d(self.x - v[0], self.y - v[1], self.z - v[2])
        else:
            return Vec3d(self.x - v, self.y - v, self.z - v)
    def __rsub__(self, v):
        if isinstance(v, Vec3d):
            return Vec3d(v.x - self.x, v.y - self.y, v.z - self.z)
        if (hasattr(v, "__getitem__")):
            return Vec3d(v[0] - self.x, v[1] - self.y, v[2] - self.z)
        else:
            return Vec3d(v - self.x, v - self.y, v - self.z)
    def __isub__(self, v):
        if isinstance(v, Vec3d):
            self.x -= v.x
            self.y -= v.y
            self.z -= v.z
        elif (hasattr(v, "__getitem__")):
            self.x -= v[0]
            self.y -= v[1]
            self.z -= v[2]
        else:
            self.x -= v
            self.y -= v
            self.z -= v
        return self

    # Multiplication
    def __mul__(self, v):
        if isinstance(v, Vec3d):
            return Vec3d(self.x*v.x, self.y*v.y, self.z*v.z)
        if (hasattr(v, "__getitem__")):
            return Vec3d(self.x*v[0], self.y*v[1], self.z*v[2])
        else:
            return Vec3d(self.x*v, self.y*v, self.z*v)
    __rmul__ = __mul__

    def __imul__(self, v):
        if isinstance(v, Vec3d):
            self.x *= v.x
            self.y *= v.y
            self.z *= v.z
        elif (hasattr(v, "__getitem__")):
            self.x *= v[0]
            self.y *= v[1]
            self.z *= v[2]
        else:
            self.x *= v
            self.y *= v
            self.z *= v
        return self

    # Division
    def __div__(self, v):
        return self._o2(v, operator.div)
    def __rdiv__(self, v):
        return self._r_o2(v, operator.div)
    def __idiv__(self, v):
        return self._io(v, operator.div)

    def __floordiv__(self, v):
        return self._o2(v, operator.floordiv)
    def __rfloordiv__(self, v):
        return self._r_o2(v, operator.floordiv)
    def __ifloordiv__(self, v):
        return self._io(v, operator.floordiv)

    def __truediv__(self, v):
        return self._o2(v, operator.truediv)
    def __rtruediv__(self, v):
        return self._r_o2(v, operator.truediv)
    def __itruediv__(self, v):
        return self._io(v, operator.floordiv)

    # Modulo
    def __mod__(self, v):
        return self._o2(v, operator.mod)
    def __rmod__(self, v):
        return self._r_o2(v, operator.mod)

    def __divmod__(self, v):
        return self._o2(v, operator.divmod)
    def __rdivmod__(self, v):
        return self._r_o2(v, operator.divmod)

    # Exponentation
    def __pow__(self, v):
        return self._o2(v, operator.pow)
    def __rpow__(self, v):
        return self._r_o2(v, operator.pow)

    # Bitwise operators
    def __lshift__(self, v):
        return self._o2(v, operator.lshift)
    def __rlshift__(self, v):
        return self._r_o2(v, operator.lshift)

    def __rshift__(self, v):
        return self._o2(v, operator.rshift)
    def __rrshift__(self, v):
        return self._r_o2(v, operator.rshift)

    def __and__(self, v):
        return self._o2(v, operator.and_)
    __rand__ = __and__

    def __or__(self, v):
        return self._o2(v, operator.or_)
    __ror__ = __or__

    def __xor__(self, v):
        return self._o2(v, operator.xor)
    __rxor__ = __xor__

    # Unary operations
    def __neg__(self):
        return Vec3d(operator.neg(self.x),
                     operator.neg(self.y),
                     operator.neg(self.z))

    def __pos__(self):
        return Vec3d(operator.pos(self.x),
                     operator.pos(self.y),
                     operator.pos(self.z))

    def __abs__(self):
        return Vec3d(abs(self.x), abs(self.y), abs(self.z))

    def __invert__(self):
        return Vec3d(-self.x, -self.y, -self.z)

    # vectory functions
    def get_length_sqrd(self):
        return self.x**2 + self.y**2 + self.z**2

    def get_length(self):
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)
    def __setlength(self, value):
        length = self.get_length()
        self.x *= value/length
        self.y *= value/length
        self.z *= value/length
    length = property(get_length, __setlength, None,
                      "gets or sets the magnitude of the vector")

    def normalized(self):
        length = self.length
        if length != 0:
            return self/length
        return Vec3d(self)

    def normalize_return_length(self):
        length = self.length
        if length != 0:
            self.x /= length
            self.y /= length
            self.z /= length
        return length

    def dot(self, v):
        return float(self.x*v[0] + self.y*v[1] + self.z*v[2])

    def get_distance(self, v):
        return math.sqrt((self.x - v[0])**2 +
                         (self.y - v[1])**2 +
                         (self.z - v[2])**2)

    def get_dist_sqrd(self, v):
        return (self.x - v[0])**2 + (self.y - v[1])**2 + (self.z - v[2])**2

    def projection(self, v):
        v_length_sqrd = v[0]*v[0] + v[1]*v[1] + v[2]*v[2]
        projected_length_times_v_length = self.dot(v)
        return v*(projected_length_times_v_length/v_length_sqrd)

    def cross(self, v):
        return Vec3d(self.y*v[2] - self.z*v[1],
                     self.z*v[0] - self.x*v[2],
                     self.x*v[1] - self.y*v[0])

    def interpolate_to(self, v, range):
        return Vec3d(self.x + (v[0] - self.x)*range,
                     self.y + (v[1] - self.y)*range,
                     self.z + (v[2] - self.z)*range)

    def convert_to_basis(self, x_vector, y_vector, z_vector):
        return Vec3d(self.dot(x_vector)/x_vector.get_length_sqrd(),
                     self.dot(y_vector)/y_vector.get_length_sqrd(),
                     self.dot(z_vector)/z_vector.get_length_sqrd())

    def __getstate__(self):
        return [self.x, self.y, self.z]

    def __setstate__(self, dict):
        self.x, self.y, self.z = dict


########################################################################
## Helper vector functions
########################################################################
def minimum(v1, v2):
    v1 = Vec3d(v1)
    v2 = Vec3d(v2)

    xmin = min(v1.x, v2.x)
    ymin = min(v1.y, v2.y)
    zmin = min(v1.z, v2.z)

    return Vec3d(xmin, ymin, zmin)


def maximum(v1, v2):
    v1 = Vec3d(v1)
    v2 = Vec3d(v2)

    xmax = max(v1.x, v2.x)
    ymax = max(v1.y, v2.y)
    zmax = max(v1.z, v2.z)

    return Vec3d(xmax, ymax, zmax)


########################################################################
## Unit Testing
########################################################################
if __name__ == "__main__":

    import unittest
    import pickle

    ####################################################################
    class UnitTestVec3d(unittest.TestCase):

        def setup(self):
            pass

        def testCreationAndAccess(self):
            v = Vec3d(111,222,333)
            self.assertTrue(v.x == 111 and v.y == 222 and v.z == 333)
            v.x = 333
            v[1] = 444
            v.z = 555
            self.assertTrue(v[0] == 333 and v[1] == 444 and v[2] == 555)

        def testMath(self):
            v = Vec3d(111,222,333)
            self.assertEqual(v + 1, Vec3d(112,223,334))
            self.assertTrue(v - 2 == [109,220,331])
            self.assertTrue(v * 3 == (333,666,999))
            self.assertTrue(v / 2.0 == Vec3d(55.5, 111, 166.5))
            self.assertTrue(v / 2 == (55, 111, 166))
            self.assertTrue(v ** Vec3d(2,3,2) == [12321, 10941048, 110889])
            self.assertTrue(v + [-11, 78, 67] == Vec3d(100, 300, 400))
            self.assertTrue(v / [11,2,9] == [10,111,37])

        def testReverseMath(self):
            v = Vec3d(111,222,333)
            self.assertTrue(1 + v == Vec3d(112,223,334))
            self.assertTrue(2 - v == [-109,-220,-331])
            self.assertTrue(3 * v == (333,666,999))
            self.assertTrue([222,999,666] / v == [2,4,2])
            self.assertTrue([111,222,333] ** Vec3d(2,3,2) ==
                            [12321, 10941048, 110889])
            self.assertTrue([-11, 78,67] + v == Vec3d(100, 300, 400))

        def testUnary(self):
            v = Vec3d(111,222,333)
            v = -v
            self.assertTrue(v == [-111,-222,-333])
            v = abs(v)
            self.assertTrue(v == [111,222,333])

        def testLength(self):
            v = Vec3d(1,4,8)
            self.assertTrue(v.length == 9)
            self.assertTrue(v.get_length_sqrd() == 81)
            self.assertTrue(v.normalize_return_length() == 9)
            self.assertTrue(v.length == 1)
            v.length = 9
            self.assertTrue(v == Vec3d(1,4,8))
            v2 = Vec3d(10, -2, 12)
            self.assertTrue(v.get_distance(v2) == (v - v2).get_length())

        def testHighLevel(self):
            basis0 = Vec3d(5.0, 0, 0)
            basis1 = Vec3d(0, .5, 0)
            basis2 = Vec3d(0, 0, 3)
            v = Vec3d(10, 1, 6)
            self.assertTrue(v.convert_to_basis(basis0, basis1, basis2) ==
                            [2, 2, 2])
            self.assertTrue(v.projection(basis0) == (10, 0, 0))
            self.assertTrue(basis0.dot(basis1) == 0)

        def testCross(self):
            lhs = Vec3d(1, .5, 3)
            rhs = Vec3d(4, 6, 1)
            self.assertTrue(lhs.cross(rhs) == [-17.5, 11, 4])

        def testComparison(self):
            int_vec = Vec3d(3, -2, 4)
            flt_vec = Vec3d(3.0, -2.0, 4.0)
            zero_vec = Vec3d(0, 0, 0)
            self.assertTrue(int_vec == flt_vec)
            self.assertTrue(int_vec != zero_vec)
            self.assertTrue((flt_vec == zero_vec) == False)
            self.assertTrue((flt_vec != int_vec) == False)
            self.assertTrue(int_vec == (3, -2, 4))
            self.assertTrue(int_vec != [0, 0, 0])
            self.assertTrue(int_vec != 5)
            self.assertTrue(int_vec != [3, -2, 4, 15])

        def testInplace(self):
            inplace_vec = Vec3d(5, 13, 17)
            inplace_ref = inplace_vec
            inplace_src = Vec3d(inplace_vec)
            inplace_vec *= .5
            inplace_vec += .5
            inplace_vec /= (3, 6, 9)
            inplace_vec += Vec3d(-1, -1, -1)
            alternate = (inplace_src*.5 + .5)/Vec3d(3, 6, 9) + [-1, -1, -1]
            self.assertEquals(inplace_vec, inplace_ref)
            self.assertEquals(inplace_vec, alternate)

        def testPickle(self):
            testvec = Vec3d(5, .3, 8.6)
            testvec_str = pickle.dumps(testvec)
            loaded_vec = pickle.loads(testvec_str)
            self.assertEquals(testvec, loaded_vec)

        def testVec3dMinMax(self):
            vmin = minimum((1,2,3),(3,2,1))
            vmax = maximum((1,2,3),(3,2,1))
            self.assertTrue(vmin == (1,2,1))
            self.assertTrue(vmax == (3,2,3))

    ####################################################################
    unittest.main()

    ####################################################################


#!/usr/bin/env python
__doc__ = """

Label transform functions.

Kisuk Lee <kisuklee@mit.edu>, 2015-2016
"""

import numpy as np
import transform
import utils

class LabelFunction(object):
    """
    Transform label.
    """

    def evaluate(self, sample, key, spec):
        if spec is not None:
            d = dict(spec)
            func = d['type']
            del d['type']
            f = globals()[func]
            f(sample, key, **d)

label_func = LabelFunction()


def binarize(sample, key, rebalancing=True):
    """Binarize label."""
    # Update sample.
    sample[key] = transform.binarize(sample[key])
    # Rebalancing.
    if rebalancing:
        wmsk = transform.rebalance_class(sample[key])
        sample[key+'_mask'] *= wmsk


def binary_class(sample, key, rebalancing=True):
    binarize(sample, key, False)
    multiclass_expansion(sample, key, [0,1], rebalancing)


def affinitize(sample, key, rebalancing=True):
    """Transfrom segmentation to 3D affinity graph."""
    affs = transform.affinitize(sample[key])
    msks = transform.affinitize_mask(sample[key+'_mask'])
    # Update sample.
    sample[key] = affs
    sample[key+'_mask'] = msks
    # Rebalancing.
    if rebalancing:
        wmsk = transform.tensor_func.rebalance_class(affs)
        sample[key+'_mask'] *= wmsk


def multiclass_expansion(sample, key, ids, rebalancing=True):
    """For semantic segmentation."""
    lbl  = sample[key]
    msk  = utils.check_volume(sample[key+'_mask'])
    lbls, msk2 = transform.multiclass_expansion(lbl, ids)
    msk *= msk2
    msks = np.tile(msk, (len(ids),1,1,1))
    # Update sample.
    sample[key] = lbls
    sample[key+'_mask'] = msks
    # Rebalancing.
    if rebalancing:
        wmsk = transform.rebalance_class(lbl)
        wmsk = np.tile(wmsk, (len(ids),1,1,1))
        sample[key+'_mask'] *= wmsk



#!/usr/bin/env python
__doc__ = """

Misalignment data augmentation.

Karan Kathpalia <karank@cs.princeton.edu>
Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import data_augmentation
import numpy as np
from utils import check_tensor

class MisalignAugment(data_augmentation.DataAugment):
    """
    Misalignment.
    """

    def __init__(self, max_trans=15.0):
        """Initialize MisalignAugment."""
        self.set_max_translation(max_trans)

    def set_max_translation(self, max_trans):
        """Set the maximum amount of translation in x and y."""
        self.MAX_TRANS = max_trans

    def prepare(self, spec, **kwargs):
        """
        TODO(kisuk): Documentation.
        """
        self.spec = spec

        # Max translation.
        if 'max_trans' in kwargs:
            self.set_max_translation(kwargs['max_trans'])

        # Random translation.
        # Always lower box is translated.
        self.x_t = int(round(self.MAX_TRANS*np.random.rand(1)))
        self.y_t = int(round(self.MAX_TRANS*np.random.rand(1)))

        # Randomly draw x/y translation independently.
        ret, pvt, zs = dict(), dict(), list()
        for k, v in self.spec.iteritems():
            z, y, x = v[-3:]
            assert z > 0
            x_dim  = self.x_t + v[-1]
            y_dim  = self.y_t + v[-2]
            ret[k] = v[:-2] + (y_dim, x_dim)
            pvt[k] = z
            zs.append(z)

        # Random direction of translation.
        x_sign = np.random.choice(['+','-'])
        y_sign = np.random.choice(['+','-'])
        self.x_t = int(eval(x_sign + str(self.x_t)))
        self.y_t = int(eval(y_sign + str(self.y_t)))

        zmin = min(zs)

        # Trivial 2D case
        if zmin == 1:
            self.do_augment = False
            ret = dict(spec)
        else:
            self.do_augment = True
            # Introduce misalignment at pivot.
            pivot = np.random.randint(1, zmin - 1)
            for k, v in pvt.iteritems():
                offset = int(v - zmin)/2  # Compute offset.
                pvt[k] = offset + pivot
            self.pivot = pvt

        return ret

    def augment(self, sample, **kwargs):
        """Apply misalignment data augmentation."""
        # DEBUG
        #print '\n[MisalignAugment]'
        #print 'misalign     z = {}'.format(self.pivot)
        #print 'misalign (x,y) = ({},{})'.format(self.x_t,self.y_t)

        ret = dict()

        if self.do_augment:
            for k, v in sample.iteritems():
                # Ensure data is 4D tensor.
                data = check_tensor(v)
                new_data = np.zeros(self.spec[k], dtype=data.dtype)
                new_data = check_tensor(new_data)
                # Dimension
                z, y, x = v.shape[-3:]
                assert z > 1
                # Copy upper box.
                xmin = max(self.x_t, 0)
                ymin = max(self.y_t, 0)
                xmax = min(self.x_t, 0) + x
                ymax = min(self.y_t, 0) + y
                pvot = self.pivot[k]
                new_data[:,0:pvot,...] = data[:,0:pvot,ymin:ymax,xmin:xmax]
                # Copy lower box.
                xmin = max(-self.x_t, 0)
                ymin = max(-self.y_t, 0)
                xmax = min(-self.x_t, 0) + x
                ymax = min(-self.y_t, 0) + y
                pvot = self.pivot[k]
                new_data[:,pvot:,...] = data[:,pvot:,ymin:ymax,xmin:xmax]
                # Augmented sample.
                ret[k] = new_data
        else:
            ret = sample

        return ret



"""
Adapted from ELEKTRONN (http://elektronn.org/).
"""

import numpy as np

cdef extern from 'warping.c':
    int fastwarp2d_opt(const float * src,
               float * dest_d,
               const int sh[3],
               const int ps[3],
               const float rot,
               const float shear,
               const float scale[2],
               const float stretch_in[2])
    int fastwarp3d_opt_zxy(const float * src,
                     float * dest_d,
                     const int sh[4],
                     const int ps[4],
                     const float rot,
                     const float shear,
                     const float scale[3],
                     const float stretch_in[4],
                     const float twist_in)


def warp2dFast(img, patch_size, rot=0, shear=0, scale=(1,1), stretch=(0,0)):
    """
    Create warped mapping for a spatial 2D input image.
    The transformation is done w.r.t to the *center* of the image.

    Parameters
    ----------

    img: array
      The array must be 3-dimensional (ch,x,y) and larger/equal the patch size
    patch_size: 2-tuple
      Patch size *excluding* channel: (px, py).
      The warping result of the input image is cropped to this size
    rot: float
      Rotation angle in deg for rotation around z-axis
    shear: float
      Shear angle in deg for shear w.r.t xy-diagonal
    scale: 2-tuple of float
      Scale per axis
    stretch: 2-tuple of float
      Fraction of perspective stretching from the center (where stretching is always 1)
      to the outer border of image per axis. The 4 entry correspond to:

      - X stretching depending on Y
      - Y stretching depending on X


    Returns
    -------

    img: np.ndarray
      Warped image (cropped to patch_size)
    """
    assert len(img.shape)==3
    rot   = rot   * np.pi / 180
    shear = shear * np.pi / 180

    scale   = np.array(scale, dtype=np.float32, order='C', ndmin=1)
    scale   = 1.0/scale
    cdef float [:] scale_view = scale
    cdef float * scale_ptr = &scale_view[0]

    stretch = np.array(stretch, dtype=np.float32, order='C', ndmin=1)
    cdef float [:] stretch_view = stretch
    cdef float * stretch_ptr = &stretch_view[0]

    img = np.ascontiguousarray(img, dtype=np.float32)
    cdef float [:, :, :] img_view = img
    cdef float * in_ptr = &img_view[0, 0, 0]

    cdef int [:] in_sh_view = np.ascontiguousarray(img.shape, dtype=np.int32)
    cdef int * in_sh_ptr = &in_sh_view[0]

    out_arr = np.zeros((img.shape[0],)+tuple(patch_size), dtype=np.float32)
    cdef float [:, :, :] out_view = out_arr
    cdef float * out_ptr = &out_view[0, 0, 0]

    cdef int [:] ps_view = np.ascontiguousarray(out_arr.shape, dtype=np.int32)
    cdef int * ps_ptr  = &ps_view[0]

    fastwarp2d_opt(in_ptr, out_ptr, in_sh_ptr, ps_ptr, rot, shear, scale_ptr, stretch_ptr)
    return out_arr


def _warp2dFastLab(lab, patch_size, img_sh, rot, shear, scale, stretch):
    rot   = rot   * np.pi / 180
    shear = shear * np.pi / 180

    scale   = np.array(scale, dtype=np.float32, order='C', ndmin=1)
    scale   = 1.0/scale
    cdef float [:] scale_view = scale
    cdef float * scale_ptr = &scale_view[0]

    stretch = np.array(stretch, dtype=np.float32, order='C', ndmin=1)
    cdef float [:] stretch_view = stretch
    cdef float * stretch_ptr = &stretch_view[0]

    new_lab = np.zeros((1,)+img_sh, dtype=np.float32)
    off = list(map(lambda x: (x[0]-x[1])//2, zip(img_sh, lab.shape)))
    new_lab[0, off[0]:lab.shape[0]+off[0], off[1]:lab.shape[1]+off[1]] = lab
    lab = new_lab
    cdef float [:, :, :] lab_view = lab
    cdef float * in_ptr = &lab_view[0, 0, 0]

    cdef int [:] in_sh_view = np.ascontiguousarray(lab.shape, dtype=np.int32)
    cdef int * in_sh_ptr = &in_sh_view[0]

    out_shape = list(map(lambda x: x[0]-2*x[1], zip(patch_size, off)))
    out_shape = (1,) + tuple(out_shape)

    out_arr = np.zeros(out_shape, dtype=np.float32)
    cdef float [:, :, :] out_view = out_arr
    cdef float * out_ptr = &out_view[0, 0, 0]

    cdef int [:] ps_view = np.ascontiguousarray(out_arr.shape, dtype=np.int32)
    cdef int * ps_ptr  = &ps_view[0]

    fastwarp2d_opt(in_ptr, out_ptr, in_sh_ptr, ps_ptr, rot, shear, scale_ptr, stretch_ptr)
    out_arr = out_arr.astype(np.int16)[0]
    return out_arr


def warp3dFast(img, patch_size, rot=0, shear=0, scale=(1,1,1), stretch=(0,0,0,0), twist=0):
    """
    Create warped mapping for a spatial 3D input image.
    The transformation is done w.r.t to the *center* of the image.

    Note that some transformations are not applied to the z-axis. This makes this function simpler
    and it is also better for anisotropic data as the different scales are not mixed up then.

    Parameters
    ----------

    img: array
      The array must be 4-dimensional (z,ch,x,y) and larger/equal the patch size
    patch_size: 3-tuple
      Patch size *excluding* channel: (pz, px, py).
      The warping result of the input image is cropped to this size
    rot: float
      Rotation angle in deg for rotation around z-axis
    shear: float
      Shear angle in deg for shear w.r.t xy-diagonal
    scale: 3-tuple of float
      Scale per axis
    stretch: 4-tuple of float
      Fraction of perspective stretching from the center (where stretching is always 1)
      to the outer border of image per axis. The 4 entry correspond to:

      - X stretching depending on Y
      - Y stretching depending on X
      - X stretching depending on Z
      - Y stretching depending on Z

    twist: float
      Dependence of the rotation angle on z in deg from center to outer border

    Returns
    -------

    img: np.ndarray
      Warped array (cropped to patch_size)

    """
    assert len(img.shape)==4

    # Rotation, shear, twist.
    rot   = rot   * np.pi / 180
    shear = shear * np.pi / 180
    twist = twist * np.pi / 180

    # Scale.
    scale = np.array(scale, dtype=np.float32, order='C', ndmin=1)
    scale = 1.0/scale
    cdef float [:] scale_view = scale
    cdef float * scale_ptr = &scale_view[0]

    # Perspective stretch.
    stretch = np.array(stretch, dtype=np.float32, order='C', ndmin=1)
    cdef float [:] stretch_view = stretch
    cdef float * stretch_ptr = &stretch_view[0]

    # Image.
    img = np.ascontiguousarray(img, dtype=np.float32)
    cdef float [:, :, :, :] img_view = img
    cdef float * in_ptr = &img_view[0, 0, 0, 0]

    # Image shape.
    cdef int [:] in_sh_view = np.ascontiguousarray(img.shape, dtype=np.int32)
    cdef int * in_sh_ptr = &in_sh_view[0]

    # Output.
    out_shape = (patch_size[0], img.shape[1], patch_size[1], patch_size[2])
    out_arr = np.zeros(out_shape, dtype=np.float32)
    cdef float [:, :, :, :] out_view = out_arr
    cdef float * out_ptr = &out_view[0, 0, 0, 0]

    # Output shape.
    cdef int [:] ps_view = np.ascontiguousarray(out_arr.shape, dtype=np.int32)
    cdef int * ps_ptr = &ps_view[0]

    fastwarp3d_opt_zxy(in_ptr, out_ptr, in_sh_ptr, ps_ptr, rot, shear,
                        scale_ptr, stretch_ptr, twist)
    return out_arr


def _warp3dFastLab(lab, patch_size, img_sh, rot, shear, scale, stretch, twist):
    n_chann = lab.shape[1]
    lab_sh  = (lab.shape[0], lab.shape[2], lab.shape[3])

    # Rotation, shear, twist.
    rot   = rot   * np.pi / 180
    shear = shear * np.pi / 180
    twist = twist * np.pi / 180

    # Scale.
    scale = np.array(scale, dtype=np.float32, order='C', ndmin=1)
    scale = 1.0/scale
    cdef float [:] scale_view = scale
    cdef float * scale_ptr = &scale_view[0]

    # Perspective stretch.
    stretch = np.array(stretch, dtype=np.float32, order='C', ndmin=1)
    cdef float [:] stretch_view = stretch
    cdef float * stretch_ptr = &stretch_view[0]

    # Label.
    new_lab_sh = (img_sh[0], n_chann, img_sh[1],img_sh[2])
    new_lab = np.zeros(new_lab_sh, dtype=np.float32)
    off = list(map(lambda x: (x[0]-x[1])//2, zip(img_sh, lab_sh)))
    new_lab[off[0]:lab_sh[0]+off[0], :, off[1]:lab_sh[1]+off[1], off[2]:lab_sh[2]+off[2]] = lab
    lab = new_lab
    cdef float [:, :, :, :] lab_view = lab
    cdef float * in_ptr = &lab_view[0, 0, 0, 0]

    # Label shape.
    cdef int [:] in_sh_view = np.ascontiguousarray(lab.shape, dtype=np.int32)
    cdef int * in_sh_ptr = &in_sh_view[0]

    out_shape = patch_size
    out_shape = (out_shape[0], n_chann, out_shape[1], out_shape[2])
    out_arr = np.zeros(out_shape, dtype=np.float32)
    cdef float [:, :, :, :] out_view = out_arr
    cdef float * out_ptr = &out_view[0, 0, 0, 0]

    # Output shape.
    cdef int [:] ps_view = np.ascontiguousarray(out_arr.shape, dtype=np.int32)
    cdef int * ps_ptr = &ps_view[0]

    fastwarp3d_opt_zxy(in_ptr, out_ptr, in_sh_ptr, ps_ptr, rot, shear,
                        scale_ptr, stretch_ptr, twist)
    # out_arr = out_arr.astype(np.int16)[:,0]
    return out_arr



#!/usr/bin/env python
__doc__ = """

DataAugmentor

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

from collections import OrderedDict
import numpy as np
from transform import *

"""
Data augmentaion pool.

Whenever adding a new data augmentation, its type name should be appended to
this list.
"""
aug_pool = ['warp','flip','grey','misalign']


class DataAugmentor(object):
    """
    Data augmentation.

    Attributes:
        _aug_list: List of data augmentation. Will be executed sequentially.
    """

    def __init__(self, spec):
        """
        TODO(kisuk): Documentation.
        """
        aug_list = []
        for s in spec:
            t = s['type']
            del s['type']
            t = t.lower()
            if t not in aug_pool:
                raise RuntimeError('unknown data augmentation type [%s]' % t)
            t = t[0].capitalize() + t[1:] + 'Augment'
            aug = eval(t + '(**s)')
            aug_list.append(aug)
        self._aug_list = aug_list

    def next_sample(self, dataset):
        raise NotImplementedError

    def random_sample(self, dataset):
        """
        TODO(kisuk): Documentation.
        """
        # Tolerant to failure due to data augmentation.
        while True:
            try:
                spec = self._prepare(dataset)
                sample, transform = dataset.random_sample(spec=spec)
                break
            except:
                pass

        # Apply data augmentation.
        for aug in self._aug_list:
            sample = aug.augment(sample, imgs=dataset.get_imgs())

        # Ensure that sample is ordered by key.
        sample = OrderedDict(sorted(sample.items(), key=lambda x: x[0]))
        return sample, transform

    def _prepare(self, dataset):
        ret = dict(dataset.get_spec())
        for aug in reversed(self._aug_list):
            ret = aug.prepare(ret, imgs=dataset.get_imgs(), **dataset.params)
        return ret


class DataAugment(object):
    """
    DataAugment interface.
    """

    def prepare(self, spec, **kwargs):
        raise NotImplementedError

    def augment(self, sample, **kwargs):
        raise NotImplementedError


class FlipAugment(DataAugment):
    """
    Random flip.
    """

    def prepare(self, spec, **kwargs):
        return dict(spec)

    def augment(self, sample, **kwargs):
        rule = np.random.rand(4) > 0.5
        return sample_func.flip(sample, rule=rule)


class GreyAugment(DataAugment):
    """
    Greyscale value augmentation.
    Randomly adjust contrast and brightness, and apply random gamma correction.
    """

    def __init__(self, mode='3D', skip_ratio=0.3):
        """
        Initialize parameters.

        Args:
            mode:
            ratio:
        """
        self.mode  = mode
        self.ratio = skip_ratio

        assert mode=='3D' or mode=='2D' or mode=='mix'

        self.CONTRAST_FACTOR   = 0.3
        self.BRIGHTNESS_FACTOR = 0.3

    def prepare(self, spec, **kwargs):
        return dict(spec)

    def augment(self, sample, **kwargs):
        #print '\n[GreyAugment]'  # DEBUG
        ret = sample
        if np.random.rand() > self.ratio:
            if self.mode == 'mix':
                mode = '3D' if np.random.rand() > 0.5 else '2D'
            else:
                mode = self.mode
            ret = eval('self.augment{}(sample, **kwargs)'.format(mode))
        return ret

    def augment2D(self, sample, **kwargs):
        """
        Adapted from ELEKTRONN (http://elektronn.org/).
        """
        #print '2D greyscale augmentation'  # DEBUG
        imgs = kwargs['imgs']
        n = len(imgs)

        # Greyscale augmentation.
        for i in range(n):
            key = imgs[i]
            for z in xrange(sample[key].shape[-3]):
                img = sample[key][...,z,:,:]
                img *= 1 + (np.random.rand() - 0.5)*self.CONTRAST_FACTOR
                img += (np.random.rand() - 0.5)*self.BRIGHTNESS_FACTOR
                img = np.clip(img, 0, 1)
                img **= 2.0**(np.random.rand()*2 - 1)
                sample[key][...,z,:,:] = img

        return sample

    def augment3D(self, sample, **kwargs):
        """
        Adapted from ELEKTRONN (http://elektronn.org/).
        """
        #print '3D greyscale augmentation'  # DEBUG

        imgs = kwargs['imgs']
        n = len(imgs)

        # Greyscale augmentation.
        for i in range(n):
            key = imgs[i]
            sample[key] *= 1 + (np.random.rand() - 0.5)*self.CONTRAST_FACTOR
            sample[key] += (np.random.rand() - 0.5)*self.BRIGHTNESS_FACTOR
            sample[key] = np.clip(sample[key], 0, 1)
            sample[key] **= 2.0**(np.random.rand()*2 - 1)

        return sample


class WarpAugment(DataAugment):
    """
    Warping.
    """

    def prepare(self, spec, **kwargs):
        pass

    def augment(self, sample, **kwargs):
        pass

"""
Whenever adding new data augmentation outside this module, it should be properly
imported as below.
"""

from misalign import MisalignAugment
from warp import WarpAugment



#!/usr/bin/env python
__doc__ = """

DaraProvider classes.

Kisuk Lee <kisuklee@mit.edu>, 2015-2016
"""

from collections import OrderedDict
import numpy as np
import parser
from dataset import VolumeDataset
from data_augmentation import DataAugmentor
from transform import *
from label_transform import *

class DataProvider(object):
    """
    DataProvider interface.
    """

    def next_sample(self):
        raise NotImplementedError

    def random_sample(self):
        raise NotImplementedError


class VolumeDataProvider(DataProvider):
    """
    DataProvider for volumetric data.

    Attributes:
        datasets: List of datasets.
        _sampling_weights: Probability of each dataset being chosen at each
                            iteration.
        _net_spec: Dictionary mapping layers' name to their input dimension.
    """

    def __init__(self, dspec_path, net_spec, params, auto_mask=True):
        """
        Initialize DataProvider.

        Args:
            dspec_path: Path to the dataset specification file.
            net_spec:   Net specification.
            params:     Various options.
            auto_mask:  Whether to automatically generate mask from
                        corresponding label.
        """
        # Params.
        drange = params['drange']            # Required.
        dprior = params.get('dprior', None)  # Optional.

        # Build Datasets.
        print '\n[VolumeDataProvider]'
        p = parser.Parser(dspec_path, net_spec, params, auto_mask=auto_mask)
        self.datasets = list()
        for d in drange:
            print 'constructing dataset %d...' % d
            config, dparams = p.parse_dataset(d)
            dataset = VolumeDataset(config, **dparams)
            self.datasets.append(dataset)

        # Sampling weight.
        self.set_sampling_weights(dprior)

        # Setup data augmentation.
        aug_spec = params.get('augment', [])  # Default is an empty list.
        self._data_aug = DataAugmentor(aug_spec)

    def set_sampling_weights(self, dprior=None):
        """
        TODO(kisuk): Documentation.
        """
        if dprior is None:
            dprior = [x.num_sample() for x in self.datasets]
        # Normalize.
        dprior = np.asarray(dprior, dtype='float32')
        dprior = dprior/np.sum(dprior)
        # Set sampling weights.
        # print 'Sampling weights: {}'.format(['%0.3f' % x for x in dprior])
        self._sampling_weights = dprior

    def next_sample(self):
        """Fetch next sample in a sample sequence."""
        return self.random_sample()  # TODO(kisuk): Temporary.

    def random_sample(self):
        """Fetch random sample."""
        # Pick one dataset randomly.
        dataset = self._get_random_dataset()
        # Draw a random sample and apply data augmenation.
        sample, transform = self._data_aug.random_sample(dataset)
        # Transform sample.
        sample = self._transform(sample, transform)
        # Ensure that sample is ordered by key.
        return OrderedDict(sorted(sample.items(), key=lambda x: x[0]))


    ####################################################################
    ## Private Helper Methods
    ####################################################################

    def _get_random_dataset(self):
        """
        Pick one dataset randomly, according to the given sampling weights.

        Returns:
            Randomly chosen dataset.
        """
        # Trivial case.
        if len(self.datasets)==1:
            return self.datasets[0]

        # Take a single experiment with a multinomial distribution, whose
        # probabilities indicate how likely each dataset be selected.
        # Output is an one-hot vector.
        sq = np.random.multinomial(1, self._sampling_weights, size=1)
        sq = np.squeeze(sq)

        # Get the index of non-zero element.
        idx = np.nonzero(sq)[0][0]

        return self.datasets[idx]

    def _transform(self, sample, transform):
        """
        TODO(kisuk): Documentation.
        """
        affinitized = False
        for key, spec in transform.iteritems():
            if spec is not None:
                if spec['type'] == 'affinitize':
                    affinitized = True
                label_func.evaluate(sample, key, spec)

        # Crop by 1 if affinitized.
        if affinitized:
            for key, data in sample.iteritems():
                sample[key] = tensor_func.crop(data, (1,1,1))

        return sample



# -*- coding: utf-8 -*-
__doc__ = """

Created on Wed Jan 28 14:27:31 2015

@author: jingpeng
"""

import numpy as np
import h5py
import tifffile

def imread(fname):
    """
    Read volumetirc data.

    Args:
        fname: Name of the file to read (hdf5 or tiff).

    Returns:
        data: Numpy 3D or 4D array.
    """
    if '.hdf5' in fname or '.h5' in fname:
        f = h5py.File(fname)
        data = np.asarray(f['/main'])
        f.close()
    elif '.tif' in fname:
        data = tifffile.imread(fname)
    else:
        raise RuntimeError('only hdf5 and tiff formats are supported')

    return data


def imsave(data, fname):
    """
    Save volumetric data.

    Args:
        data: Numpy array to save.
        fname: Name of the file to save (hdf5 or tiff).
    """
    if '.hdf5' in fname or '.h5' in fname:
        f = h5py.File(fname)
        f.create_dataset('/main', data=data)
        f.close()
    elif '.tif' in fname:
        tifffile.imsave(fname, data)
    else:
        raise RuntimeError('only hdf5 and tiff formats are supported')



#!/usr/bin/env python
__doc__ = """

ConfigData classes.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import copy
import numpy as np

import emio
from tensor import TensorData
from transform import *
from utils import *

class ConfigData(TensorData):
    """
    ConfigData class.
    """

    def __init__(self, config, section):
        """
        Build data from a ConfiParser object generated by Parser's
        parse_dataset method.
        """
        # Preprocessing.
        data, fov, offset = self._prepare_data(config, section)

        # Initialize TensorData.
        super(ConfigData, self).__init__(data, fov=fov, offset=offset)

    ####################################################################
    ## Private Helper Methods
    ####################################################################

    def _prepare_data(self, config, section):
        """Prepare data from config."""
        assert config.has_section(section)

        # Either read data from the specified file, or generate data with the
        # specified shape and filler.
        if config.has_option(section, 'file'):
            data = emio.imread(config.get(section, 'file'))
        elif config.has_option(section, 'shape'):
            shape = config.get(section, 'shape')
            # Ensure that shape is tuple.
            shape = tuple(eval(str(shape)))
            if config.has_option(section, 'filler'):
                filler = eval(config.get(section, 'filler'))
            else:
                filler = {'type':'zero'}
            data = fill_data(shape, filler=filler)
        else:
            raise RuntimeError('Invalid data section [%s].' % section)

        # Field of View (patch size).
        if config.has_option(section, 'fov'):
            fov = config.get(section, 'fov')
            # Ensure that fov is tuple.
            fov = tuple(eval(str(fov)))
        else:
            fov = (0,0,0)

        # Offset.
        if config.has_option(section, 'offset'):
            offset = config.get(section, 'offset')
            # Ensure that offset is tuple.
            offset = tuple(eval(str(offset)))
        else:
            offset = (0,0,0)

        # List of global preprocessing.
        if config.has_option(section, 'preprocess'):
            preprocess = config.get(section, 'preprocess').split('\n')
            preprocess = [eval(x) for x in preprocess]
        else:
            preprocess = list()

        # Check validity of each preprocessing.
        for pp in preprocess:
            assert isinstance(pp, dict)
            assert 'type' in pp

        # Perform preprocessing seequentially.
        data = check_tensor(data)
        for pp in preprocess:
            data = tensor_func.evaluate(data, pp)

        return data, fov, offset


class ConfigLabel(ConfigData):
    """
    ConfigLabel class.
    """

    def __init__(self, config, section):
        """Build data from config."""
        # Initialize ConfigData.
        super(ConfigLabel, self).__init__(config, section)

        # Transformation
        self._transformation(config, section)

    def get_transform(self):
        return copy.deepcopy(self._transform)

    ####################################################################
    ## Private Helper Methods
    ####################################################################

    def _transformation(self, config, section):
        """
        TODO(kisuk): Documentation.
        """
        # List of local transformation.
        if config.has_option(section, 'transform'):
            transform = eval(config.get(section, 'transform'))
        else:
            transform = None

        # Check the validity of each transformation.
        if transform is not None:
            assert isinstance(transform, dict)
            assert 'type' in transform

        self._transform = transform



#!/usr/bin/env python
__doc__ = """

Functions for data transformation.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

from collections import OrderedDict
import numpy as np
from utils import *
from vector import Vec3d, minimum, maximum

def transform_tensor(func, data, *args, **kwargs):
    """Apply func to each channel of data (4D tensor)."""
    data = check_tensor(data)
    arrs = list()
    f = globals()[func]
    for c in xrange(data.shape[0]):
        vol = f(data[c,...], *args, **kwargs)
        arrs.append(check_tensor(vol))
    return np.concatenate(arrs, axis=0)


class SampleFunction(object):
    """
    Apply 3D volume function to every volume in a data sample.
    """

    def __getattr__(self, name):
        def transform_fn(sample, *args, **kwargs):
            return self._transform_sample(name, sample, *args, **kwargs)
        return transform_fn

    def _transform_sample(self, func, sample, *args, **kwargs):
        """Apply func to a sample."""
        ret = OrderedDict()
        for key, data in sample.iteritems():
            ret[key] = transform_tensor(func, data, *args, **kwargs)
        return ret


class TensorFunction(object):
    """
    Apply 3D volume function to each channel of 4D tensor.
    """

    def __getattr__(self, func):
        def transform_fn(data, *args, **kwargs):
            return transform_tensor(func, data, *args, **kwargs)
        return transform_fn

    def evaluate(self, data, spec):
        """Evaluate tensor function specified by spec on data."""
        d = dict(spec)
        func = d['type']
        del d['type']
        return transform_tensor(func, data, **d)


sample_func = SampleFunction()
tensor_func = TensorFunction()


def crop(img, offset=(0,0,0), size=None):
    """
    TODO(kisuk): Documentation.
    """
    img = check_volume(img)
    if size is None:
        size = tuple(Vec3d(img.shape) - Vec3d(offset))
    ret = np.zeros(size, dtype=img.dtype)
    v1  = Vec3d(offset)
    v2  = v1 + Vec3d(size)
    ret[:] = img[v1[0]:v2[0],v1[1]:v2[1],v1[2]:v2[2]]
    return ret

####################################################################
## Preprocessing
####################################################################

def mirror_border(img, fov):
    """
    TODO(kisuk) Documentation.
    """
    img = check_volume(img)

    # Validate FoV
    fov = np.asarray(fov)
    fov = Vec3d(fov.astype('uint32'))

    # Pad size
    top = fov / 2
    btm = fov - top - (1,1,1)
    pad_with = [(top[0],btm[0]),(top[1],btm[1]),(top[2],btm[2])]
    # TODO(kisuk): Should we force an odd-sized fov?

    # TODO(kisuk): 'symmetric' or 'reflect'?
    return np.pad(img, pad_with, mode='reflect')


def standardize(img, mode='2D', dtype='float32'):
    """Standard normalization (zero mean, unit standard deviation)."""
    img = check_volume(img)
    ret = np.zeros(img.shape, dtype=dtype)

    # Standardize function (zero mean & unit standard deviation).
    f = lambda x: (x - np.mean(x)) / np.std(x)

    if mode == '2D':
        for z in xrange(img.shape[0]):
            ret[z,:,:] = f(img[z,:,:])
    elif mode == '3D':
        ret[:] = f(img)
    else:
        raise RuntimeError("mode must be either '2D' or '3D'")

    return ret


def rescale(img, min_val=0.0, max_val=1.0, dtype='float32'):
    """Rescale."""
    img = check_volume(img)
    ret = np.zeros(img.shape, dtype=dtype)

    # Rescale to [min_val, max_val].
    ret[:] = img
    ret -= np.min(ret)
    ret /= np.ptp(ret)
    ret *= (max_val - min_val)
    ret += min_val

    return ret


def divideby(img, val=255.0, dtype='float32'):
    """Rescale."""
    img = check_volume(img)
    ret = np.zeros(img.shape, dtype=dtype)

    ret[:] = img
    ret /= val

    return ret


####################################################################
## Data Augmentations
####################################################################

def flip(data, rule):
    """Flip data according to a specified rule.

    Args:
        data: 3D numpy array to be transformed.
        rule: Transform rule, specified as a Boolean array.
             [z reflection,
              y reflection,
              x reflection,
              xy transpose]

    Returns:
        data: Transformed data.
    """
    data = check_tensor(data)

    assert np.size(rule)==4

    # z reflection
    if rule[0]:
        data = data[:,::-1,:,:]
    # y reflection
    if rule[1]:
        data = data[:,:,::-1,:]
    # x reflection
    if rule[2]:
        data = data[:,:,:,::-1]
    # Transpose in xy.
    if rule[3]:
        data = data.transpose(0,1,3,2)

    return data


def revert_flip(data, rule, dst=None):
    """
    TODO(kisuk): Documentation.
    """
    data = check_tensor(data)

    assert np.size(rule)==4

    # Special treat for affinity.
    is_affinity = False if dst is None else True
    if is_affinity:
        (dz,dy,dx) = dst
        assert data.shape[-4]==3
        assert dx and abs(dx) < data.shape[-1]
        assert dy and abs(dy) < data.shape[-2]
        assert dz and abs(dz) < data.shape[-3]

    # Transpose in xy.
    if rule[3]:
        data = data.transpose(0,1,3,2)
        # Swap x/y-affinity maps.
        if is_affinity:
            data[[0,1],...] = data[[1,0],...]
    # x reflection
    if rule[2]:
        data = data[:,:,:,::-1]
        # Special treatment for x-affinity.
        if is_affinity:
            if dx > 0:
                data[0,:,:,dx:] = data[0,:,:,:-dx]
                data[0,:,:,:dx].fill(0)
            else:
                dx = abs(dx)
                data[0,:,:,:-dx] = data[0,:,:,dx:]
                data[0,:,:,-dx:].fill(0)
    # y reflection
    if rule[1]:
        data = data[:,:,::-1,:]
        # Special treatment for y-affinity.
        if is_affinity:
            if dy > 0:
                data[1,:,dy:,:] = data[1,:,:-dy,:]
                data[1,:,:dy,:].fill(0)
            else:
                dy = abs(dy)
                data[1,:,:-dy,:] = data[1,:,dy:,:]
                data[1,:,-dy:,:].fill(0)
    # z reflection
    if rule[0]:
        data = data[:,::-1,:,:]
        # Special treatment for z-affinity.
        if is_affinity:
            if dz > 0:
                data[2,dz:,:,:] = data[2,:-dz,:,:]
                data[2,:dz,:,:].fill(0)
            else:
                dz = abs(dz)
                data[2,:-dz,:,:] = data[2,dz:,:,:]
                data[2,-dz:,:,:].fill(0)

    return data


####################################################################
## Label Transformations
####################################################################

def binarize(img, dtype='float32'):
    """Binarize image.

    Normally used to turn a ground truth segmentation into a ground truth
    boundary map, binary representation for each voxel being neuronal boundary
    or not.

    Args:
        img: 3D indexed image, with each index corresponding to each segment.

    Returns:
        ret: Binarized image.
    """
    img = check_volume(img)
    ret = np.zeros(img.shape, dtype=dtype)
    ret[:] = (img>0).astype(dtype)
    return ret


def multiclass_expansion(img, ids, dtype='float32'):
    """Expand an indexed image to one-hot representation for multiclass
    classification.

    Args:
        img: 3D indexed image, with each index corresponding to each class.
        ids: a list of class IDs to expand.

    Returns:
        ret: an expanded 4D tensor.
        msk:
    """
    img = check_volume(img)
    ret = np.zeros((len(ids),) + img.shape, dtype=dtype)
    msk = np.zeros(img.shape, dtype=dtype)
    for i, l in enumerate(ids):
        idx = (img == l)
        msk[idx] = 1
        ret[i,...] = idx.astype(dtype)
    return ret, msk


def binary_class(img, dtype='float32'):
    """
    TODO(kisuk): Documentation.
    """
    img = check_volume(img)
    img = binarize(img, dtype=dtype)
    ret, _ = multiclass_expansion(img, ids=[0,1], dtype=dtype)
    return ret


def affinitize(img, dst=(1,1,1), dtype='float32'):
    """
    Transform segmentation to 3D affinity graph.

    Args:
        img: 3D indexed image, with each index corresponding to each segment.

    Returns:
        ret: 3D affinity graph (4D tensor), 3 channels for z, y, x direction.
    """
    img = check_volume(img)
    ret = np.zeros((3,) + img.shape, dtype=dtype)

    (dz,dy,dx) = dst

    # z-affinity.
    assert dz and abs(dz) < img.shape[-3]
    if dz > 0:
        ret[2,dz:,:,:] = (img[dz:,:,:]==img[:-dz,:,:]) & (img[dz:,:,:]>0)
    else:
        dz = abs(dz)
        ret[2,:-dz,:,:] = (img[dz:,:,:]==img[:-dz,:,:]) & (img[dz:,:,:]>0)

    # y-affinity.
    assert dy and abs(dy) < img.shape[-2]
    if dy > 0:
        ret[1,:,dy:,:] = (img[:,dy:,:]==img[:,:-dy,:]) & (img[:,dy:,:]>0)
    else:
        dy = abs(dy)
        ret[1,:,:-dy,:] = (img[:,dy:,:]==img[:,:-dy,:]) & (img[:,dy:,:]>0)

    # x-affinity.
    assert dx and abs(dx) < img.shape[-1]
    if dx > 0:
        ret[0,:,:,dx:] = (img[:,:,dx:]==img[:,:,:-dx]) & (img[:,:,dx:]>0)
    else:
        dx = abs(dx)
        ret[0,:,:,:-dx] = (img[:,:,dx:]==img[:,:,:-dx]) & (img[:,:,dx:]>0)

    return ret

####################################################################
## Mask Transformations
####################################################################

def affinitize_mask(msk, dst=(1,1,1), dtype='float32'):
    """
    Transform binary mask to affinity mask.

    Args:
        msk: 3D binary mask.

    Returns:
        ret: 3D affinity mask (4D tensor), 3 channels for z, y, x direction.
    """
    msk = check_volume(msk)
    ret = np.zeros((3,) + msk.shape, dtype=dtype)

    (dz,dy,dx) = dst

    # z-affinity.
    assert dz and abs(dz) < msk.shape[-3]
    if dz > 0:
        ret[2,dz:,:,:] = (msk[dz:,:,:]>0) | (msk[:-dz,:,:]>0)
    else:
        dz = abs(dz)
        ret[2,:-dz,:,:] = (msk[dz:,:,:]>0) | (msk[:-dz,:,:]>0)

    # y-affinity.
    assert dy and abs(dy) < msk.shape[-2]
    if dy > 0:
        ret[1,:,dy:,:] = (msk[:,dy:,:]>0) | (msk[:,:-dy,:]>0)
    else:
        dy = abs(dy)
        ret[1,:,:-dy,:] = (msk[:,dy:,:]>0) | (msk[:,:-dy,:]>0)

    # x-affinity.
    assert dx and abs(dx) < msk.shape[-1]
    if dx > 0:
        ret[0,:,:,dx:] = (msk[:,:,dx:]>0) | (msk[:,:,:-dx]>0)
    else:
        dx = abs(dx)
        ret[0,:,:,:-dx] = (msk[:,:,dx:]>0) | (msk[:,:,:-dx]>0)

    return ret

####################################################################
## Rebalancing
####################################################################

def rebalance_class(img, msk=None, dtype='float32'):
    """Class-rebalancing."""
    img = check_volume(img)
    ret = np.zeros(img.shape, dtype=dtype)

    masked = img if msk is None else img[msk>0]
    unique_lbl, num_lbls = np.unique(masked, return_counts=True)
    assert(len(num_lbls)>0)

    if len(num_lbls)==1:
        # TODO(kisuk):
        #   This is to make rebalancing exactly the same as in ZNNv1 and v4,
        #   but not sure about how reasonable this value (0.5) is, and about
        #   if this can also be applied to multiclass case (e.g. semantic
        #   segmentation).
        ret[:] = 0.5
    else:
        weights = 1.0/num_lbls
        weights = weights/np.sum(weights)
        for idx, lbl in enumerate(unique_lbl):
            ret[img==lbl] = weights[idx]

    return ret


# def rebalance_class(img, dtype='float32'):
#     """Class-rebalancing."""
#     img = check_volume(img)
#     ret = np.zeros(img.shape, dtype=dtype)
#
#     _, idx, cnt = np.unique(img, return_inverse=True, return_counts=True)
#     assert(len(cnt)>0)
#
#     if len(cnt)==1:
#         # TODO(kisuk):
#         #   This is to make rebalancing exactly the same as in ZNNv1 and v4,
#         #   but not sure about how reasonable this value (0.5) is, and about
#         #   if this can also be applied to multiclass case (e.g. semantic
#         #   segmentation).
#         ret[:] = 0.5
#     else:
#         weights = 1.0/cnt
#         weights = weights/np.sum(weights)
#         ret = weights[idx.reshape(ret.shape)]
#
#     return ret

########################################################################
## Unit Testing
########################################################################
if __name__ == "__main__":

    import unittest

    ####################################################################
    class UnitTestTransform(unittest.TestCase):

        def setup(self):
            pass

        def testCrop(self):
            img = np.random.rand(4,4,4)
            a = crop(img, (3,3,3))
            b = img[:-1,:-1,:-1]
            self.assertTrue(np.array_equal(a,b))
            a = crop(img, (3,3,3), (1,1,1))
            b = img[1:,1:,1:]
            self.assertTrue(np.array_equal(a,b))

    ####################################################################
    unittest.main()

    ####################################################################



#!/usr/bin/env python
import h5py

from data_provider import VolumeDataProvider
import time
from vector import Vec3d

if __name__ == "__main__":

    # Data spec path.
    dspec_path = 'test_spec/pinky.spec'

    # Net specification.
    # fov   = Vec3d(9,109,109)
    # outsz = Vec3d(10,100,100)
    # insz  = outsz + fov - Vec3d(1,1,1)
    # net_spec = dict(input=tuple(insz), label=tuple(outsz))
    net_spec = dict(input=(1,208,208), label=(1,100,100))

    # Parameters.
    params = dict()
    params['drange']  = [0,1,2,3,4]
    params['dprior']  = None
    params['border']  = None
    params['augment'] = [dict(type='flip')]

    # Create VolumeDataProvider.
    dp = VolumeDataProvider(dspec_path, net_spec, params)

    # for i in range(10000):
    #     start = time.time()
    #     sample = dp.random_sample()
    #     print time.time() - start

    # Dump a single randome sample.
    sample = dp.random_sample()
    print 'Save as file...'
    f = h5py.File('sample.h5')
    for name, data in sample.iteritems():
        f.create_dataset('/' + name, data=data)
    f.close()

    # # Dump the whole dataset.
    # print 'Save as file...'
    # dataset = dp.datasets[0]._data  # Illegal access. Don't try this at home.
    # f = h5py.File('dataset.h5')
    # for name, data in dataset.iteritems():
    #     f.create_dataset('/' + name, data=data.get_data())
    # f.close()



#!/usr/bin/env python
__doc__ = """

Dataset classes.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

from collections import OrderedDict
import copy
import numpy as np

from box import Box
from config_data import ConfigData, ConfigLabel
from vector import Vec3d

class Dataset(object):
    """
    Dataset interface.
    """

    def next_sample(self):
        raise NotImplementedError

    def random_sample(self):
        raise NotImplementedError


class VolumeDataset(Dataset):
    """
    Dataset for volumetric data.

    Attributes:
        _data:  Dictionary mapping layer's name to TensorData, which contains
                4D volumetric data (e.g. EM image stacks, label stacks, etc.).

        _label: List of label layer's name.

        _spec:  Net specification. Dictionary mapping layer's name to its input
                dimension (either 3-tuple or 4-tuple).

        _range: Range of valid coordinates for accessing data given a net spec.
                It depends both on data and net specs.

        params: Dataset-specific parameters.
    """

    def __init__(self, config, **kwargs):
        """Initialize VolumeDataset."""
        self.build_from_config(config)
        # Set dataset-specific params.
        self.params = dict()
        for k, v in kwargs.iteritems():
            self.params[k] = v

    def __getattr__(self, name):
        """Access dataset-specific params."""
        assert name in self.params
        return self.params[name]

    def build_from_config(self, config):
        """
        Build dataset from a ConfiParser object generated by Parser's
        parse_dataset method.
        """
        self._reset()

        # First pass for images and labels.
        assert config.has_section('dataset')
        for name, data in config.items('dataset'):
            assert config.has_section(data)
            if '_mask' in data:
                # Mask will be processed later.
                continue
            if 'label' in data:
                self._data[name] = ConfigLabel(config, data)
                self._label.append(name)
            else:
                self._data[name] = ConfigData(config, data)
                self._image.append(name)

        # Second pass for masks.
        for name, data in config.items('dataset'):
            if '_mask' in data:
                if config.has_option(data, 'shape'):
                    # Lazy filling of mask shape. Since the shape of mask should
                    # be the same as the shape of corresponding label, it can be
                    # known only after having processed label in the first pass.
                    label, _ = data.split('_mask')
                    shape = self._data[label].shape()
                    config.set(data, 'shape', shape)
                self._data[name] = ConfigData(config, data)

        # Set dataset spec.
        spec = dict()
        for name, data in self._data.iteritems():
            spec[name] = tuple(data.fov())
        self.set_spec(spec)

    def get_spec(self):
        """Return dataset spec."""
        # TODO(kisuk):
        #   spec's value type is tuple, which is immutable. Do we still need to
        #   deepcopy it?
        return copy.deepcopy(self._spec)

    def get_imgs(self):
        return copy.deepcopy(self._image)

    def set_spec(self, spec):
        """Set spec and update valid range."""
        # Order by key
        self._spec = OrderedDict(sorted(spec.items(), key=lambda x: x[0]))
        self._update_range()

    def num_sample(self):
        """Return number of samples in valid range."""
        s = self._range.size()
        return s[0]*s[1]*s[2]

    def get_range(self):
        """Return valid range."""
        return Box(self._range)

    def get_sample(self, pos):
        """Extract a sample centered on pos.

        Every data in the sample is guaranteed to be center-aligned.

        Args:
            pos: Center coordinate of the sample.

        Returns:
            sample:     Dictionary mapping input layer's name to data.
            transform:  Dictionary mapping label layer's name to the type of
                        label transformation specified by user.

        """
        # self._spec is guaranteed to be ordered by key, so using OrderedDict
        # and iterating over already-sorted self._spec together guarantee the
        # sample is sorted.
        sample = OrderedDict()
        for name in self._spec.keys():
            sample[name] = self._data[name].get_patch(pos)

        transform = dict()
        for name in self._label:
            transform[name] = self._data[name].get_transform()

        return sample, transform

    def next_sample(self, spec=None):
        """Fetch next sample in a sample sequence."""
        return self.random_sample(spec)  # Currently just pick randomly.

    def random_sample(self, spec=None):
        """Fetch sample randomly"""
        # Dynamically change spec.
        if spec is not None:
            original_spec = self._spec
            self.set_spec(spec)

        pos = self._random_location()
        ret = self.get_sample(pos)

        # Return to original spec.
        if spec is not None:
            self.set_spec(original_spec)

        # ret is a 2-tuple (sample, transform).
        return ret

    ####################################################################
    ## Private Helper Methods
    ####################################################################

    def _reset(self):
        """Reset all attributes."""
        self._data  = dict()
        self._image = list()
        self._label = list()
        self._spec  = None
        self._range = None

    def _random_location(self):
        """Return one of the valid locations randomly."""
        s = self._range.size()
        z = np.random.randint(0, s[0])
        y = np.random.randint(0, s[1])
        x = np.random.randint(0, s[2])
        # Global coordinate system.
        return Vec3d(z,y,x) + self._range.min()
        # DEBUG
        #return self._range.min()

    def _update_range(self):
        """
        Update valid range. It's computed by intersecting the valid range of
        each TensorData.
        """
        self._range = None
        for name, dim in self._spec.iteritems():
            # Update patch size.
            self._data[name].set_fov(dim[-3:])
            # Update valid range.
            r = self._data[name].range()
            if self._range is None:
                self._range = r
            else:
                self._range = self._range.intersect(r)



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import time
import numpy as np

# try:
#     from ._warping import warp2dFast, warp3dFast, _warp2dFastLab, _warp3dFastLab
# except ImportError:
#     raise RuntimeError('_warping.so Cython extension not found.\n'
#                        'Please run setup.py or manually cythonize _warping.pyx.')
from _warping import warp2dFast, warp3dFast, _warp2dFastLab, _warp3dFastLab


def warp2dJoint(img, lab, patch_size, rot, shear, scale, stretch):
    """
    Warp image and label data jointly. Non-image labels are ignored i.e. lab must be 3d to be warped

    Parameters
    ----------

    img: array
      Image data
      The array must be 3-dimensional (ch,x,y) and larger/equal the patch size
    lab: array
      Label data (with offsets subtracted)
    patch_size: 2-tuple
      Patch size *excluding* channel for the image: (px, py).
      The warping result of the input image is cropped to this size
    rot: float
      Rotation angle in deg for rotation around z-axis
    shear: float
      Shear angle in deg for shear w.r.t xy-diagonal
    scale: 3-tuple of float
      Scale per axis
    stretch: 4-tuple of float
      Fraction of perspective stretching from the center (where stretching is always 1)
      to the outer border of image per axis. The 4 entry correspond to:

      - X stretching depending on Y
      - Y stretching depending on X

    Returns
    -------

    img, lab: np.ndarrays
      Warped image and labels (cropped to patch_size)

    """
    if len(lab.shape) == 2:
        lab = _warp2dFastLab(lab, patch_size, img.shape[1:], rot, shear, scale, stretch)

    img = warp2dFast(img, patch_size, rot, shear, scale, stretch)
    return img, lab


def warp3dJoint(img, lab, patch_size, rot=0, shear=0, scale=(1, 1, 1), stretch=(0, 0, 0, 0), twist=0):
    """
    Warp image and label data jointly. Non-image labels are ignored i.e. lab must be 3d to be warped

    Parameters
    ----------

    img: array
      Image data
      The array must be 4-dimensional (z,ch,x,y) and larger/equal the patch size
    lab: array
      Label data (with offsets subtracted)
    patch_size: 3-tuple
      Patch size *excluding* channel for the image: (pz, px, py).
      The warping result of the input image is cropped to this size
    rot: float
      Rotation angle in deg for rotation around z-axis
    shear: float
      Shear angle in deg for shear w.r.t xy-diagonal
    scale: 3-tuple of float
      Scale per axis
    stretch: 4-tuple of float
      Fraction of perspective stretching from the center (where stretching is always 1)
      to the outer border of image per axis. The 4 entry correspond to:

      - X stretching depending on Y
      - Y stretching depending on X
      - X stretching depending on Z
      - Y stretching depending on Z

    twist: float
      Dependence of the rotation angle on z in deg from center to outer border

    Returns
    -------

    img, lab: np.ndarrays
      Warped image and labels (cropped to patch_size)

    """
    if len(lab.shape) == 3:
        lab = _warp3dFastLab(lab, patch_size, np.array(img.shape)[[0, 2, 3]], rot, shear, scale, stretch, twist)

    img = warp3dFast(img, patch_size, rot, shear, scale, stretch, twist)
    return img, lab


def warp3d(img, patch_size, rot=0, shear=0, scale=(1, 1, 1), stretch=(0, 0, 0, 0), twist=0):
    return warp3dFast(img, patch_size, rot, shear, scale, stretch, twist)

def warp3dLab(lab, patch_size, size, rot=0, shear=0, scale=(1, 1, 1), stretch=(0, 0, 0, 0), twist=0):
    return _warp3dFastLab(lab, patch_size, size, rot, shear, scale, stretch, twist)

### Utilities #################################################################
###############################################################################


def getCornerIx(sh):
    """Returns array-indices of corner elements for n-dim shape"""

    def getGrayCode(n, n_dim):
        if n == 0:
            return np.zeros(n_dim, dtype=np.int)
        return np.array([(n // 2**i) % 2 for i in range(max(n_dim, int(np.ceil(np.log2(n)))))])

    sh = np.array(sh) - 1  ###TODO
    n_dim = len(sh)
    ix = []
    for i in xrange(2**n_dim):
        ix.append(getGrayCode(i, n_dim))

    ix = np.array(ix)
    corners = ix * sh
    return corners


def _warpCorners2d(sh, corners, rot=0, shear=0, scale=(1, 1), stretch=(0, 0), plot=False):
    """
    Create warped coordinates of corners
    """
    rot = rot * np.pi / 180
    shear = shear * np.pi / 180
    scale = np.array(scale)
    scale = 1.0 / scale
    stretch = np.array(stretch)
    corners = corners.astype(np.float).copy()

    x_center_off = float(sh[0]) / 2 - 0.5
    y_center_off = float(sh[1]) / 2 - 0.5

    stretch[0] /= x_center_off
    stretch[1] /= y_center_off

    x = corners[:, 0] - x_center_off
    y = corners[:, 1] - y_center_off

    xt = x * (scale[0] + stretch[0] * y)
    yt = y * (scale[1] + stretch[1] * x)
    u = xt * np.cos(rot - shear) - yt * np.sin(rot + shear) + x_center_off
    v = yt * np.cos(rot + shear) + xt * np.sin(rot - shear) + y_center_off

    if plot:
        coords = np.array([u, v]).T
        coords = coords[[3, 2, 0, 1, 3]]
        plt.figure(figsize=(5, 5))
        plt.scatter(corners[:, 1], corners[:, 0], c='b')
        plt.plot(corners[:, 1], corners[:, 0], 'b:')
        plt.scatter(coords[:, 1], coords[:, 0], c='r', marker='x')
        plt.plot(coords[:, 1], coords[:, 0], c='r')
        plt.axes().set_aspect('equal')
        plt.gca().invert_yaxis()
        plt.grid()

    return np.array([u, v]).T


def _warpCorners3d(sh, corners, rot=0, shear=0, scale=(1, 1, 1), stretch=(0, 0, 0, 0), twist=0):
    """
    Create warped coordinates of corners
    """
    rot = rot * np.pi / 180
    shear = shear * np.pi / 180
    twist = twist * np.pi / 180
    scale = np.array(scale)
    scale = 1.0 / scale
    stretch = np.array(stretch)
    corners = corners.astype(np.float).copy()

    z_center_off = float(sh[0]) / 2 - 0.5
    x_center_off = float(sh[1]) / 2 - 0.5
    y_center_off = float(sh[2]) / 2 - 0.5

    stretch[0] /= x_center_off
    stretch[1] /= y_center_off
    stretch[2] /= z_center_off
    stretch[3] /= z_center_off
    twist /= z_center_off

    z = corners[:, 0] - z_center_off
    x = corners[:, 1] - x_center_off
    y = corners[:, 2] - y_center_off

    w = z * scale[2] + z_center_off
    rot = rot + (z * twist)

    xt = x * (scale[0] + stretch[0] * y + stretch[2] * z)
    yt = y * (scale[1] + stretch[1] * x + stretch[3] * z)
    u = xt * np.cos(rot - shear) - yt * np.sin(rot + shear) + x_center_off
    v = yt * np.cos(rot + shear) + xt * np.sin(rot - shear) + y_center_off

    return np.array((w, u, v)).T


def getRequiredPatchSize(patch_size, rot, shear, scale, stretch, twist=None):
    """
    Given desired patch size and warping parameters:
    return required size for warping input patch
    """
    patch_size = np.array(patch_size)
    corners = getCornerIx(patch_size)

    if len(patch_size) == 2:
        coords = _warpCorners2d(patch_size, corners, rot, shear, scale, stretch)
    elif len(patch_size) == 3:
        coords = _warpCorners3d(patch_size, corners, rot, shear, scale, stretch, twist)

    eff_size = np.ceil(coords.max(axis=0) - coords.min(axis=0))  # effective range
    left_exc = np.floor(np.abs(np.minimum(coords.min(axis=0), 0)))  # how much image needs to be added left
    right_exc = np.ceil(np.maximum(coords.max(axis=0) - patch_size + 1, 0))
    total_exc = np.maximum(left_exc, right_exc)  # how much image must be added centrally
    req_size = patch_size + 2 * total_exc

    return req_size.astype(np.int), eff_size.astype(np.int), left_exc.astype(np.int)


def getWarpParams(patch_size, amount=1.0, **kwargs):
    """
    To be called from CNNData. Get warping parameters + required warping input patch size.
    """
    if amount > 1:
        print 'WARNING: warpAugment amount > 1 this requires more than 1.4 bigger patches before warping'
    rot_max = 15 * amount
    shear_max = 3 * amount
    scale_max = 1.2 * amount
    stretch_max = 0.1 * amount
    n_dim = len(patch_size)

    # Data-specific max.
    if 'scale_max' in kwargs:
       scale_max = kwargs['scale_max']

    shear = shear_max * 2 * (np.random.rand() - 0.5)
    if n_dim == 3:
        twist = rot_max * 2 * (np.random.rand() - 0.5)
        rot = min(rot_max - abs(twist), rot_max * (np.random.rand()))
        scale = 1 - (scale_max - 1) * np.random.rand()
        scale = (scale, scale, 1)
        stretch = stretch_max * 2 * (np.random.rand(4) - 0.5)
    elif n_dim == 2:
        rot = rot_max * 2 * (np.random.rand() - 0.5)
        scale = 1 - (scale_max - 1) * np.random.rand(2)
        stretch = stretch_max * 2 * (np.random.rand(2) - 0.5)
        twist = None

    # DEBUG
    if 'rot' in kwargs:  # 0
       rot = kwargs['rot']
    if 'shear' in kwargs:  # 0
       shear = kwargs['shear']
    if 'scale' in kwargs:  # (1,1,1)
       scale = kwargs['scale']
    if 'stretch' in kwargs:  # (1,1,1,1)
       stretch = kwargs['stretch']
    if 'twist' in kwargs:  # 0
       twist = kwargs['twist']

    req_size, _, _ = getRequiredPatchSize(patch_size, rot, shear, scale,
                                          stretch, twist)

    return req_size, rot, shear, scale, stretch, twist


def test():
    try:
        img_s = np.random.rand(11, 11)
        img_s = np.concatenate((img_s[None], np.exp(img_s[None])), axis=0)
        out = warp2dFast(img_s, (11, 11), 0, 0, (1, 1), (0.0, 0.0))
    except Exception as e:
        print """%s
        Warping is broken. Most likeley the distributed _warping.so is not binary compatible to your system.""" % (e, )


test()

##############################################################################################################

#def paddImage(img,  ext_size, left_exc):
#  new_img = np.ones(ext_size, dtype=img.dtype)
#  xs, ys = img.shape
#  xo, yo = left_exc
#  new_img[xo:xo+xs, yo:yo+ys] = img
#
#  return new_img


def maketestimage(sh):
    img = np.ones(sh) * 0.5
    xs, ys = sh
    try:
        d = np.diag(np.ones(xs))
        img[:xs, :xs] += d
        img[:xs, :xs] += d[::-1]
        img[-xs:, -xs:] += d
        img[-xs:, -xs:] += d[::-1]
    except:
        d = np.diag(np.ones(ys))
        img[:ys, :ys] += d
        img[:ys, :ys] += d[::-1]
        img[-ys:, -ys:] += d
        img[-ys:, -ys:] += d[::-1]

    img[0, :] += 1
    img[:, 0] += 1
    img[-1, :] += 1
    img[:, -1] += 1
    if sh[0] > 80:
        img[30, :] += 1
        img[:, 30] += 1
        img[-31, :] += 1
        img[:, -31] += 1
    return img / img.max()


if __name__ == "__main__":
    #  test_img = io.imread('Lichtenstein.png')
    #  test_img = test_img.mean(axis=2)
    #  s1 = test_img.shape[0]
    #  s2 = test_img.shape[1]
    ps = (200, 200)

    if True:
        ext_size, rot, shear, scale, stretch, twist = getWarpParams(ps,
                                                                    amount=1.0)
        t = []
        for i in xrange(10000):
            ext_size, rot, shear, scale, stretch, twist = getWarpParams(ps, amount=1.0); t.append(ext_size)
    #  img_in = maketestimage(eff_size)
    #  img_in = paddImage(img_in, ext_size, left_exc)[None]
        img_in = maketestimage(ext_size)[None]
        out = warp2dFast(img_in, ps, rot, shear, scale, stretch)

        plt.figure()
        plt.subplot(121)
        plt.imshow(img_in[0], interpolation='none', cmap='gray')
        plt.hlines(ext_size[0] / 2 - 0.5, 0, ext_size[1] - 1, color='r')
        plt.vlines(ext_size[1] / 2 - 0.5, 0, ext_size[0] - 1, color='r')
        plt.subplot(122)
        plt.imshow(out[0], interpolation='none', cmap='gray')
        plt.hlines(ps[0] / 2 - 0.5, 0, ps[1] - 1, color='r')
        plt.vlines(ps[1] / 2 - 0.5, 0, ps[0] - 1, color='r')

    if False:  # visual 2d
        #out  = _warp2d_c(test_img, 20, 10, (1,1.1), (0.1, 0))
        test_img = np.concatenate((test_img[None], np.exp(test_img[None])), axis=0)
        out2 = warp2dFast(test_img, (512, 512), 20, 10, (1, 1.1), (0.1, 0))
        plt.figure()
        plt.subplot(121)
        plt.imshow(test_img, interpolation='none', cmap='gray')
        plt.subplot(122)
        plt.imshow(out2[0], interpolation='none', cmap='gray')

    if False:
        img_s = maketestimage((11, 11))
        img_s = np.concatenate((img_s[None], np.exp(img_s[None])), axis=0)
        out = warp2dFast(img_s, (11, 11), 0, 0, (1, 1), (0.0, 0.0))

    if False:
        img_s = maketestimage((110, 110))
        img_s = np.concatenate((img_s[None], ) * 4, axis=0)
        img_s = np.concatenate((img_s[None], np.exp(img_s[None])), axis=0)
        out = warp3dFast(img_s, (4, 110, 110), 0, 0, (1, 1, 1),
                         (0.0, 0.0, 0.0, 0.0), 10)

    if False:  # visual 3d
        n = 100

        img_s = np.tile(test_img, n)
        img_s = img_s.reshape((s1, n, s2))
        img_s = np.swapaxes(img_s, 1, 0)

        patch_size = img_s.shape
        off = 0
        lab = img_s[off:-off, off:-off, off:-off]

        img = np.concatenate((test_img[None], np.exp(test_img[None])), axis=0)

        img1, lab1 = warpAugment(img_s[None], lab, patch_size=patch_size)

        for i in xrange(n):
            plt.imsave('/tmp/%i-img.png' % i, img1[0, i, :, :] / 255)

        for i in xrange(lab1.shape[0]):
            plt.imsave('/tmp/%i-lab.png' % (i + off), lab1[i, :, :])

    if False:  # visual 3d
        n = 40
        img_s = np.tile(test_img, n)
        img_s = img_s.reshape((s1, n, s2))
        img_s = np.swapaxes(img_s, 1, 2)

        wow1 = warp3dFast(img_s[None], (s1, s2, n), 0, 0, (1, 1, 1),
                          (0.1, 0.1, 0.1, -0.1), 10)
        for i in xrange(n):
            plt.imsave('/tmp/%i-ref.png' % i, wow1[:, :, i] / 255)

        wow2 = _warp3dFastLab(img_s[20:-20, 20:-20], (s1 - 40, s2 - 40, n),
                              (s1, s2, n), 0, 0, (1, 1, 1), (0.1, 0.1, 0.1, -0.1), 10)
        for i in xrange(wow2.shape[2]):
            plt.imsave('/tmp/%i.png' % i, wow2[:, :, i] / 255)

    if False:  # 3d timing
        s = 400
        test = np.random.rand(s, s, s).astype(np.float32)
        test2 = np.random.rand(s * 2, s * 2, s * 2).astype(np.float32)
        t0 = time.time()
        wow1 = warp3dFast(test[None], (s, s, s), 20, 5, (1, 1, 1), (0.1, 0.1, 0.1, 0.1), 10)
        #wow1 = warp3dFast(test[None], (s,s,s))
        print time.time() - t0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from Cython.Build import cythonize

from setuptools import setup, Extension

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    'Cython>=0.22',
    'numpy>=1.10',
    'scipy>=0.9'
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='dataprovider',
    version='0.1.6',
    description="Deep learning platform-independent volumetric data provider for training 3D convolutional network.",
    long_description=readme + '\n\n' + history,
    author="Seung Lab",
    author_email='tartavull@gmail.com',
    url='https://github.com/tartavull/dataprovider',
    packages=[
        'dataprovider',
    ],
    package_dir={'dataprovider':
                 'dataprovider'},
    entry_points={
        'console_scripts': [
            'dataprovider=dataprovider.cli:main'
        ]
    },
    include_package_data=True,
    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.xml', '*.html', '*.js','*.c','*.pyx']
    },
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='dataprovider',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    ext_modules=cythonize([Extension('*', ['dataprovider/warping/*.pyx'])])
)



#!/usr/bin/env python
__doc__ = """

DataSpecParser class.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import ConfigParser

from vector import Vec3d

class Parser(object):
    """
    Parser class.
    """

    def __init__(self, dspec_path, net_spec, params, auto_mask=True):
        """Initialize a Parser object.

        Args:
            dspec_path: Data spec path.
            net_spec: Net spec, a dictionary containing layer-data name pairs.
            params: Parameter dictionary.
        """
        # Construct a ConfigParser object.
        config = ConfigParser.ConfigParser()
        config.read(dspec_path)

        # Set attributes.
        self._config   = config
        self.net_spec  = net_spec
        self.params    = params
        self.auto_mask = auto_mask

        # Set dataset-specific params.
        self.dparams = self._parse_dataset_params(config)


    def parse_dataset(self, dataset_id):
        """
        TODO(kisuk): Documentation.

        Args:
            dataset_id:

        Returns:
            dataset: ConfigParser object containing dataset info.
        """
        config = ConfigParser.ConfigParser()

        section = 'dataset'
        if self._config.has_section(section):
            config.add_section(section)
            # for name, data in self._config.items(section):
            for name in self.net_spec.keys():
                if self._config.has_option(section, name):
                    data = self._config.get(section, name)
                    config.set(section, name, data)
                    self.parse_data(config, name, data, dataset_id)
        else:
            raise RuntimeError('dataset section does not exist.')

        # Treat special case of affinity.
        self._treat_affinity(config)
        # Treat border.
        self._treat_border(config)

        # Dataset-specific params.
        dparams = dict()
        dparams['dataset_id'] = dataset_id
        for k, v in self.dparams.iteritems():
            dparams[k] = v[dataset_id]

        return config, dparams

    def parse_data(self, config, name, data, idx):
        """
        TODO(kisuk): Documentation.

        Args:
            config:
            name:
            data:
            idx:
        """
        if self._config.has_section(data):
            config.add_section(data)
            for option, value in self._config.items(data):
                config.set(data, option, value)
        else:
            raise RuntimeError('data section [%s] does not exist.' % data)

        # File
        if config.has_option(data, 'file'):
            value = config.get(data, 'file')
            assert self._config.has_option('files', value)
            flist = self._config.get('files', value).split('\n')
            config.set(data, 'file', flist[idx])

        # FoV
        fov = self.net_spec[name][-3:]
        config.set(data, 'fov', fov)

        # Offset
        if not config.has_option(data, 'offset'):
            config.set(data, 'offset', (0,0,0))

        # Add mask if data is label.
        if self.auto_mask and 'label' in data:
            self._add_mask(config, name, data, idx)

    ####################################################################
    ## Private Helper Methods
    ####################################################################

    def _parse_dataset_params(self, config):
        """
        TODO(kisuk): Documentation.
        """
        ret = dict()
        if config.has_section('params'):
            opts = config.options('params')
            for opt in opts:
                ret[opt] = eval(config.get('params',opt))
        return ret

    def _add_mask(self, config, name, data, idx):
        """Add mask for label.

        Each label is supposed to have corresponding mask, an equal-sized
        volume with postive real values. During training, cost and gradient
        volumes are element-wise multiplied by mask.

        Args:
            config: ConfigParser object containing dataset info.
            name: Layer name for label.
            data: Section name for label data.
            idx: Dataset ID.
        """
        assert config.has_section(data)

        # Mask name is assumed to be label name with '_mask' postfix.
        name = name + '_mask'
        mask = data + '_mask'
        config.set('dataset', name, mask)

        # Add mask data.
        config.add_section(mask)
        for option, value in config.items(data):
            if option == 'file':
                continue
            elif option == 'mask':
                # Use [mask] option of label, if exists, to instantiate
                # [file] option of mask.
                assert self._config.has_option('files', value)
                flist  = self._config.get('files', value).split('\n')
                option = 'file'
                value  = flist[idx]
            config.set(mask, option, value)

        # Add default mask filler.
        # If mask is not built from file, build it from [shape] and [filler]
        # options. [shape] option should be added later dynamically.
        if not config.has_option(mask, 'file'):
            config.set(mask, 'shape', '(z,y,x)')  # Place holder
            config.set(mask, 'filler', "{'type':'one'}")

    def _is_affinity(self, config, data):
        """Check if data is affinity."""
        ret = False
        if config.has_section(data):
            if config.has_option(data, 'transform'):
                tf = config.get(data, 'transform')
                if 'affinitize' in tf:
                    ret = True
        return ret

    def _has_affinity(self, config):
        """Check if dataset contains affinity data."""
        ret = False
        for _, data in config.items('dataset'):
            if self._is_affinity(config, data):
                ret = True
                break
        return ret

    def _treat_affinity(self, config):
        """
        TODO(kisuk): Documentation.
        """
        if self._has_affinity(config):
            for _, data in config.items('dataset'):
                assert config.has_section(data)
                assert config.has_option(data, 'fov')
                # Increment FoV by 1.
                fov = config.get(data, 'fov')
                fov = tuple(x+1 for x in fov)
                config.set(data, 'fov', fov)

    def _treat_border(self, config):
        """
        TODO(kisuk): Documentation.
        """
        border_func = self.params.get('border', None)
        if border_func is not None:
            if border_func['type'] is 'mirror_border':
                for _, data in config.items('dataset'):
                    # Apply only to images.
                    if not 'image' in data:
                        continue
                    assert config.has_section(data)
                    assert config.has_option(data, 'offset')
                    offset = config.get(data, 'offset')
                    # Append border mirroring to the preprocessing list.
                    if config.has_option(data, 'preprocess'):
                        pp = config.get(data, 'Preprocess') + '\n'
                    else:
                        pp = ''
                    pp += str(border_func)
                    config.set(data, 'preprocess', pp)
                    # Update offset.
                    fov = border_func['fov']
                    offset = Vec3d(offset) - Vec3d(fov)/2
                    config.set(data, 'offset', tuple(offset))
            else:
                msg = 'unknown border mode [%s].' % border_func['type']
                raise RuntimeError(msg)


if __name__ == "__main__":

    # Data spec path.
    dspec_path = 'test_spec/zfish.spec'

    # FoV of the net.
    fov = (9,109,109)

    # For training.
    net_spec = dict(input=(18,208,208), label=(10,100,100))
    params = dict()
    params['border']  = dict(type='mirror_border', fov=fov)
    params['augment'] = [dict(type='flip')]
    params['drange']  = [0]

    # For inference.
    # net_spec = dict(input=(18,208,208))
    # params = dict()
    # params['border']  = dict(type='mirror_border', fov=fov)
    # params['drange']  = [0]

    # Parser
    p = Parser(dspec_path, net_spec, params)
    config, dparams = p.parse_dataset(0)
    assert dparams['dataset_id']==0
    print dparams
    f = open('zfish_dataset0.spec', 'w')
    config.write(f)
    f.close()



#!/usr/bin/env python
__doc__ = """

3D Box class.

Box -- two vectors, forming a box

Modified the code from
https://wiki.python.org/moin/PointsAndRectangles

Kisuk Lee <kisuklee@mit.edu>, 2015-2016
"""

import math

from vector import Vec3d, minimum, maximum

class Box(object):
    """
    A 3D box identified by two vectors.

    set_coords       -- reset box coordinates
    contains         -- is a point inside?
    overlaps         -- does a box overlap?
    intersect        -- intersection between two boxes, if exist
    merge            -- merge two boxes
    min              -- get min corner point vector
    max              -- get max corner point vector
    expand_by        -- in-place grow (or shrink)
    expanded_by      -- grow (or shrink)

    """

    def __init__(self, v1_or_box, v2=None):
        """Initialize a box from another box or two vectors."""
        if v2 == None:
            self.set_coords(v1_or_box.min(), v1_or_box.max())
        else:
            self.set_coords(Vec3d(v1_or_box), Vec3d(v2))

    def set_coords(self, v1, v2):
        """Reset the box coordinates."""
        self._min  = minimum(v1, v2)
        self._max  = maximum(v1, v2)
        self._size = self._max - self._min

    def size(self):
        return Vec3d(self._size)

    def contains(self, v):
        """Return true if a point is inside the box."""
        (x,y,z) = Vec3d(v)
        return (self._min.x <= x < self._max.x and
                self._min.y <= y < self._max.y and
                self._min.z <= z < self._max.z)

    def overlaps(self, other):
        """Return true if a box overlaps this box."""
        return (self._max.x > other._min.x and self._min.x < other._max.x and
                self._max.y > other._min.y and self._min.y < other._max.y and
                self._max.z > other._min.z and self._min.z < other._max.z)

    def intersect(self, other):
        """Return intersection between this and other box, if overlaps."""
        if self.overlaps(other):
            # min/max corners
            vmin = maximum(self._min, other._min)
            vmax = minimum(self._max, other._max)

            return Box(vmin,vmax)
        else:
            return None

    def merge(self, other):
        """
        Return merge of this and other box. Two boxes need not overlap.
        """
        # min/max corners
        vmin = minimum(self._min, other._min)
        vmax = maximum(self._max, other._max)

        return Box(vmin,vmax)

    def translate(self, v):
        self._min += v
        self._max += v

    def min(self):
        """Return the min corner as a vector."""
        return Vec3d(self._min)

    def max(self):
        """Return the max corner as a vector."""
        return Vec3d(self._max)

    def expand_by(self, v):
        """In-place expansion by v."""
        self._min -= v
        self._max += v
        assert(self._min.x < self._max.x and
               self._min.y < self._max.y and
               self._min.z < self._max.z)

    def expanded_by(self, v):
        """Return a box with extended borders."""
        v1 = self._min - v
        v2 = self._max + v
        assert v1.x < v2.x and v1.y < v2.y and v1.z < v2.z
        return Box(v1,v2)

    # Comparison
    def __eq__(self, b):
        return self._min == b.min() and self._max == b.max()

    def __ne__(self, b):
        return not(self == b)

    # String representaion (for printing and debugging)
    def __str__( self ):
        return "<Box (%s)-(%s)>"%(self._min,self._max)

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__,self._min,self._max)


# helper box functions
def centered_box(c, s):
    """Return a box of size s centered on c."""
    center = Vec3d(c)
    size   = Vec3d(s)
    half   = size/2
    assert size.x >= 0 and size.y >= 0 and size.z >= 0
    v1 = center - half
    v2 = v1 + size
    return Box(v1,v2)


########################################################################
## Unit Testing
########################################################################
if __name__ == "__main__":

    import unittest

    ####################################################################
    class UnitTestBox(unittest.TestCase):

        def setup(self):
            pass

        def testCreationAndAccess(self):
            b = Box((1,1,1),[2,2,2])
            self.assertTrue(b.min()  == Vec3d(1,1,1))
            self.assertTrue(b.max()  == Vec3d(2,2,2))
            self.assertTrue(b.size() == Vec3d(1,1,1))
            b = Box((1,0,1),[0,1,0])
            self.assertTrue(b.min()  == Vec3d(0,0,0))
            self.assertTrue(b.max()  == Vec3d(1,1,1))
            self.assertTrue(b.size() == Vec3d(1,1,1))
            b1 = Box(b)
            self.assertTrue(b1.min()  == Vec3d(0,0,0))
            self.assertTrue(b1.max()  == Vec3d(1,1,1))
            self.assertTrue(b1.size() == Vec3d(1,1,1))

        def testComparison(self):
            b1 = Box((1,1,1),[2,2,2])
            b2 = Box((2,2,2),[1,1,1])
            self.assertTrue(b1 == b2)
            b3 = Box((1,1,1),[3,3,3])
            self.assertTrue(b1 != b3)

        def testTranslate(self):
            b = Box((1,1,1),[2,2,2])
            b.translate((-1,-1,-1))
            self.assertTrue(b == Box((0,0,0),(1,1,1)))

        def testContains(self):
            b = Box(Vec3d(1,1,1),[3,3,3])
            self.assertTrue(b.contains(b.min()))
            self.assertTrue(not b.contains(b.max()))
            self.assertTrue(b.contains((2,2,2)))
            self.assertTrue(not b.contains((-2,-2,-2)))

        def testOverlaps(self):
            b1 = Box(Vec3d(1,1,1),[2,2,2])
            b2 = Box(Vec3d(3,3,3),[2,2,2])
            self.assertTrue(b1.overlaps(b1) and b2.overlaps(b2))
            self.assertTrue((not b1.overlaps(b2)) and (not b2.overlaps(b1)))
            b2.translate((-0.5,-0.5,-0.5))
            self.assertTrue(b1.overlaps(b2) and b2.overlaps(b1))

        def testIntersect(self):
            b1 = Box(Vec3d(1,1,1),[2,2,2])
            b2 = Box(Vec3d(3,3,3),[2,2,2])
            self.assertTrue(b1.intersect(b2) == None)
            b2.translate((-0.5,-0.5,-0.5))
            b3 = b1.intersect(b2)
            self.assertTrue(b3 == Box((1.5,1.5,1.5),(2,2,2)))

        def testMerge(self):
            b1 = Box(Vec3d(1,1,1),[2,2,2])
            b2 = Box(Vec3d(3,3,3),[2,2,2])
            b3 = b1.merge(b2)
            self.assertTrue(b3 == Box((1,1,1),(3,3,3)))
            b4 = b3.merge(Box((0,0,0),(-1,-1,-1)))
            self.assertTrue(b4 == Box((-1,-1,-1),(3,3,3)))

        def testExpand(self):
            b1 = Box(Vec3d(1,1,1),[2,2,2])
            b1.expand_by(1)
            self.assertTrue(b1 == Box((0,0,0),(3,3,3)))
            b1.expand_by((-1,0,0))
            self.assertTrue(b1 == Box((1,0,0),(2,3,3)))

        def testCenteredBox(self):
            b = centered_box((0,0,0),(2,2,2))
            self.assertTrue(b == Box((-1,-1,-1),(1,1,1)))
            b = centered_box((0,0,0),(3,3,3))
            self.assertTrue(b == Box((-1,-1,-1),(2,2,2)))

    ####################################################################
    unittest.main()

    ####################################################################


#!/usr/bin/env python
__doc__ = """

ForwardScanner.

Kisuk Lee <kisuklee@mit.edu>, 2016
"""

import numpy as np
import math

import blend
from box import Box, centered_box
from tensor import WritableTensorData as WTD, WritableTensorDataWithMask as WTDM
from vector import *

class ForwardScanner(object):
    """
    ForwardScanner.

    Attributes:
        dataset:
        scan_spec:
        params:
    """

    def __init__(self, dataset, scan_spec, params=None):
        """
        Initialize ForwardScanner.
        """
        self._init()

        self.dataset   = dataset
        self.scan_spec = scan_spec
        self.params    = params if params is not None else dict()

        self._setup()

    def pull(self):
        """
        TODO(kisuk): Documentation.
        """
        ret = None
        if self.counter < len(self.locs):
            assert self.current is None
            idx = self.counter
            loc = self.locs[idx]
            print '({}/{}) loc: {}'.format(idx+1, len(self.locs), tuple(loc))
            ret, _ = self.dataset.get_sample(loc)
            self.current = loc
            self.counter += 1
        return ret

    def push(self, sample):
        """
        TODO(kisuk): Documentation

        Args:
            sample:
            kwargs:
        """
        assert self.current is not None
        self.outputs.push(self.current, sample)
        self.current = None

    ####################################################################
    ## Private Methods.
    ####################################################################

    def _init(self):
        """Initialize all attributes."""
        self.dataset        = None
        self.scan_spec      = dict()
        self.params         = dict()
        self.offset         = (0,0,0)
        self.stride         = (0,0,0)
        self.grid           = (0,0,0)
        self.vmin           = None
        self.vmax           = None
        self.default_stride = None
        self.coords         = [None]*3
        self.locs           = None
        self.counter        = 0
        self.current        = None
        self.outputs        = None

    def _setup(self):
        """
        TODO(kisuk): Documentation.
        """
        self.offset = Vec3d(self.params.get('offset', (0,0,0)))
        self.stride = Vec3d(self.params.get('stride', (0,0,0)))
        self.grid   = Vec3d(self.params.get('grid',   (0,0,0)))

        # TODO(kisuk): Validity check?

        self.vmin = self.dataset.get_range().min() + self.offset
        self.vmax = self.dataset.get_range().max()

        # TODO(kisuk): Validity check?

        # Order is important!
        self._setup_stride()
        self._setup_coords()
        self._prepare_outputs()

    def _setup_stride(self):
        """
        TODO(kisuk): Documentation.
        """
        stride = None
        for k, v in self.scan_spec.iteritems():
            box = centered_box(Vec3d(0,0,0), v[-3:])
            if stride is None:
                stride = box
            else:
                stride = stride.intersect(box)
        self.default_stride = stride.size()

    def _setup_coords(self):
        """
        TODO(kisuk): Documentation.
        """
        self._setup_coord(0)  # z-dimension
        self._setup_coord(1)  # y-dimension
        self._setup_coord(2)  # x-dimension

        self.locs = list()
        for z in self.coords[0]:
            for y in self.coords[1]:
                for x in self.coords[2]:
                    self.locs.append(Vec3d(z,y,x))

    def _setup_coord(self, dim):
        """
        TODO(kisuk): Documenatation.

        Args:
            dim: 0: z-dimension.
                 1: y-dimension.
                 2: x-dimension.
        """
        assert dim < 3

        # min & max coordinates.
        cmin = int(self.vmin[dim])
        cmax = int(self.vmax[dim])
        assert cmin < cmax

        # Dimension-specific params.
        stride = self.stride[dim]
        grid   = int(self.grid[dim])
        coord  = set()

        # Non-overlapping stride.
        if stride == 0:
            stride = self.default_stride[dim]
        # Overlapping stride given by an overlapping ratio.
        elif stride > 0 and stride < 1:
            stride = math.ceil(stride * self.default_stride[dim])
        self.stride[dim] = int(stride)
        stride = self.stride[dim]

        # Automatic full spanning.
        if grid == 0:
            grid = (cmax - cmin - 1)/stride + 1
            coord.add(cmax-1)  # Offcut

        # Scan coordinates.
        for i in range(grid):
            c = cmin + i*stride
            if c >= cmax:
                break
            coord.add(c)

        # Sanity check.
        assert cmin+(grid-1)*stride < cmax

        # Sort coords.
        self.coords[dim] = sorted(coord)

    def _prepare_outputs(self):
        """Prepare outputs according to the blending mode."""
        # Inference with overlapping window.
        diff = self.stride - self.default_stride
        overlap = True if diff[0]<0 or diff[1]<0 or diff[2]<0 else False
        # Prepare outputs.
        blend_mode = self.params.get('blend', '')
        self.outputs = blend.prepare_outputs(self.scan_spec, self.locs,
                                    blend=overlap, blend_mode=blend_mode)

if __name__ == "__main__":

    import data_provider

    # Data spec path
    dspec_path = 'test_spec/piriform.spec'

    # Net specification
    net_spec = {}
    net_spec['input'] = (18,208,208)

    # Parameters
    params = {}
    params['border']  = 'mirror'
    params['drange']  = [1]

    # VolumeDataProvider
    dp = VolumeDataProvider(dspec_path, net_spec, params)
